"""Cluster handlers module for Zigbee Home Automation."""

from __future__ import annotations

from collections.abc import Awaitable, Callable, Coroutine, Iterator
import contextlib
from dataclasses import dataclass
from enum import Enum
import functools
import logging
from typing import TYPE_CHECKING, Any, Final, ParamSpec, TypedDict

import zigpy.exceptions
from zigpy.typing import UNDEFINED, UndefinedType
import zigpy.util
import zigpy.zcl
from zigpy.zcl import (
    AttributeReadEvent,
    AttributeReportedEvent,
    AttributeUpdatedEvent,
    AttributeWrittenEvent,
)
from zigpy.zcl.foundation import (
    CommandSchema,
    ConfigureReportingResponseRecord,
    Status,
    ZCLAttributeDef,
)

from zha.application.const import (
    ZHA_CLUSTER_HANDLER_MSG,
    ZHA_CLUSTER_HANDLER_MSG_BIND,
    ZHA_CLUSTER_HANDLER_MSG_CFG_RPT,
)
from zha.application.helpers import safe_read
from zha.event import EventBase
from zha.exceptions import ZHAException
from zha.mixins import LogMixin
from zha.zigbee.cluster_handlers.const import (
    ARGS,
    ATTRIBUTE_ID,
    ATTRIBUTE_NAME,
    ATTRIBUTE_VALUE,
    CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
    CLUSTER_HANDLER_EVENT,
    CLUSTER_HANDLER_ZDO,
    CLUSTER_ID,
    CLUSTER_READS_PER_REQ,
    COMMAND,
    PARAMS,
    REPORT_CONFIG_ATTR_PER_REQ,
    SIGNAL_ATTR_UPDATED,
    UNIQUE_ID,
    VALUE,
)

if TYPE_CHECKING:
    from zha.zigbee.endpoint import Endpoint

_LOGGER = logging.getLogger(__name__)
RETRYABLE_REQUEST_DECORATOR = zigpy.util.retryable_request(tries=3)
UNPROXIED_CLUSTER_METHODS = {"general_command"}


_P = ParamSpec("_P")
_FuncType = Callable[_P, Awaitable[Any]]
_ReturnFuncType = Callable[_P, Coroutine[Any, Any, Any]]


@contextlib.contextmanager
def wrap_zigpy_exceptions() -> Iterator[None]:
    """Wrap zigpy exceptions in `ZHAException` exceptions."""
    try:
        yield
    except TimeoutError as exc:
        raise ZHAException("Failed to send request: device did not respond") from exc
    except zigpy.exceptions.ZigbeeException as exc:
        message = "Failed to send request"

        if str(exc):
            message = f"{message}: {exc}"

        raise ZHAException(message) from exc


def retry_request[**P](
    func: Callable[P, Awaitable[Any]],
) -> Callable[P, Coroutine[Any, Any, Any]]:
    """Send a request with retries and wrap expected zigpy exceptions."""

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        with wrap_zigpy_exceptions():
            return await RETRYABLE_REQUEST_DECORATOR(func)(*args, **kwargs)

    return wrapper


class AttrReportConfig(TypedDict, total=True):
    """Configuration to report for the attributes."""

    # An attribute name
    attr: str
    # The config for the attribute reporting configuration consists of a tuple for
    # (minimum_reported_time_interval_s, maximum_reported_time_interval_s, value_delta)
    config: tuple[int, int, int | float]


def parse_and_log_command(cluster_handler, tsn, command_id, args):
    """Parse and log a zigbee cluster command."""
    try:
        name = cluster_handler.cluster.server_commands[command_id].name
    except KeyError:
        name = f"0x{command_id:02X}"

    cluster_handler.debug(
        "received '%s' command with %s args on cluster_id '%s' tsn '%s'",
        name,
        args,
        cluster_handler.cluster.cluster_id,
        tsn,
    )
    return name


class ClusterHandlerStatus(Enum):
    """Status of a cluster handler."""

    CREATED = 1
    CONFIGURED = 2
    INITIALIZED = 3


@dataclass(kw_only=True, frozen=True)
class ClusterAttributeUpdatedEvent:
    """Event to signal that a cluster attribute has been updated."""

    attribute_id: int
    attribute_name: str
    attribute_value: Any
    cluster_handler_unique_id: str
    cluster_id: int
    event_type: Final[str] = CLUSTER_HANDLER_EVENT
    event: Final[str] = CLUSTER_HANDLER_ATTRIBUTE_UPDATED


@dataclass(kw_only=True, frozen=True)
class ClusterBindEvent:
    """Event generated when the cluster is bound."""

    cluster_name: str
    cluster_id: int
    success: bool
    cluster_handler_unique_id: str
    event_type: Final[str] = ZHA_CLUSTER_HANDLER_MSG
    event: Final[str] = ZHA_CLUSTER_HANDLER_MSG_BIND


@dataclass(kw_only=True, frozen=True)
class ClusterConfigureReportingEvent:
    """Event generates when a cluster configures attribute reporting."""

    cluster_name: str
    cluster_id: int
    attributes: dict[str, dict[str, Any]]
    cluster_handler_unique_id: str
    event_type: Final[str] = ZHA_CLUSTER_HANDLER_MSG
    event: Final[str] = ZHA_CLUSTER_HANDLER_MSG_CFG_RPT


@dataclass(kw_only=True, frozen=True)
class ClusterInfo:
    """Cluster information."""

    id: int
    name: str
    type: str
    commands: dict[int, str]


@dataclass(kw_only=True, frozen=True)
class ClusterHandlerInfo:
    """Cluster handler information."""

    class_name: str
    generic_id: str
    endpoint_id: str
    cluster: ClusterInfo
    id: str
    unique_id: str
    status: ClusterHandlerStatus
    value_attribute: str | None = None


class ClusterHandler(LogMixin, EventBase):
    """Base cluster handler for a Zigbee cluster."""

    REPORT_CONFIG: tuple[AttrReportConfig, ...] = ()
    BIND: bool = True

    # Dict of attributes to read on cluster handler initialization.
    # Dict keys -- attribute ID or names, with bool value indicating whether a cached
    # attribute read is acceptable.
    ZCL_INIT_ATTRS: dict[str, bool] = {}

    def __init__(self, cluster: zigpy.zcl.Cluster, endpoint: Endpoint) -> None:
        """Initialize ClusterHandler."""
        super().__init__()
        self._generic_id = f"cluster_handler_0x{cluster.cluster_id:04x}"
        self._endpoint: Endpoint = endpoint
        self._cluster: zigpy.zcl.Cluster = cluster
        self._id: str = f"{endpoint.id}:0x{cluster.cluster_id:04x}"
        unique_id: str = endpoint.unique_id.replace("-", ":")
        self._unique_id: str = f"{unique_id}:0x{cluster.cluster_id:04x}"
        if not hasattr(self, "_value_attribute") and self.REPORT_CONFIG:
            attr_def: ZCLAttributeDef = self.cluster.attributes_by_name[
                self.REPORT_CONFIG[0]["attr"]
            ]
            self.value_attribute = attr_def.name
        self._status: ClusterHandlerStatus = ClusterHandlerStatus.CREATED
        self.data_cache: dict[str, Any] = {}
        self._unsubs: list[Callable[[], None]] = []

    def on_add(self) -> None:
        """Call when cluster handler is added."""
        self._cluster.add_listener(self)
        for event_type in (
            AttributeReadEvent,
            AttributeReportedEvent,
            AttributeUpdatedEvent,
            AttributeWrittenEvent,
        ):
            self._unsubs.append(
                self._cluster.on_event(
                    event_type.event_type, self._handle_attribute_updated_event
                )
            )

    def on_remove(self) -> None:
        """Call when cluster handler will be removed."""
        self._cluster.remove_listener(self)
        for unsub in self._unsubs:
            unsub()
        self._unsubs.clear()

    def _handle_attribute_updated_event(
        self,
        event: AttributeReadEvent
        | AttributeReportedEvent
        | AttributeUpdatedEvent
        | AttributeWrittenEvent,
    ) -> None:
        """Handle attribute updated event from zigpy."""
        self.debug(
            "cluster_handler[%s] attribute_updated - cluster[%s] attr[%s] value[%s]",
            self.name,
            self.cluster.name,
            event.attribute_name,
            event.value,
        )
        self.emit(
            CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
            ClusterAttributeUpdatedEvent(
                attribute_id=event.attribute_id,
                attribute_name=event.attribute_name,
                attribute_value=event.value,
                cluster_handler_unique_id=self.unique_id,
                cluster_id=self.cluster.cluster_id,
            ),
        )

    @classmethod
    def matches(cls, cluster: zigpy.zcl.Cluster, endpoint: Endpoint) -> bool:  # pylint: disable=unused-argument
        """Filter the cluster match for specific devices."""
        return True

    @functools.cached_property
    def info_object(self) -> ClusterHandlerInfo:
        """Return info about this cluster handler."""
        return ClusterHandlerInfo(
            class_name=self.__class__.__name__,
            generic_id=self._generic_id,
            endpoint_id=self._endpoint.id,
            cluster=ClusterInfo(
                id=self._cluster.cluster_id,
                name=self._cluster.name,
                type="client" if self._cluster.is_client else "server",
                commands=self._cluster.commands,
            ),
            id=self._id,
            unique_id=self._unique_id,
            status=self._status.name,
            value_attribute=getattr(self, "value_attribute", None),
        )

    @functools.cached_property
    def id(self) -> str:
        """Return cluster handler id unique for this device only."""
        return self._id

    @functools.cached_property
    def generic_id(self) -> str:
        """Return the generic id for this cluster handler."""
        return self._generic_id

    @functools.cached_property
    def unique_id(self) -> str:
        """Return the unique id for this cluster handler."""
        return self._unique_id

    @functools.cached_property
    def cluster(self) -> zigpy.zcl.Cluster:
        """Return the zigpy cluster for this cluster handler."""
        return self._cluster

    @functools.cached_property
    def name(self) -> str:
        """Return friendly name."""
        return self.cluster.ep_attribute or self._generic_id

    @property
    def status(self) -> ClusterHandlerStatus:
        """Return the status of the cluster handler."""
        return self._status

    def __hash__(self) -> int:
        """Make this a hashable."""
        return hash(self._unique_id)

    async def bind(self) -> None:
        """Bind a zigbee cluster.

        This also swallows ZigbeeException exceptions that are thrown when
        devices are unreachable.
        """
        try:
            res = await RETRYABLE_REQUEST_DECORATOR(self.cluster.bind)()
            self.debug("bound '%s' cluster: %s", self.cluster.ep_attribute, res[0])
            self._endpoint.device.emit(
                ZHA_CLUSTER_HANDLER_MSG_BIND,
                ClusterBindEvent(
                    cluster_name=self.cluster.name,
                    cluster_id=self.cluster.cluster_id,
                    cluster_handler_unique_id=self.unique_id,
                    success=res[0] == 0,
                ),
            )
        except (zigpy.exceptions.ZigbeeException, TimeoutError) as ex:
            self.debug(
                "Failed to bind '%s' cluster: %s",
                self.cluster.ep_attribute,
                str(ex),
                exc_info=ex,
            )
            self._endpoint.device.emit(
                ZHA_CLUSTER_HANDLER_MSG_BIND,
                ClusterBindEvent(
                    cluster_name=self.cluster.name,
                    cluster_id=self.cluster.cluster_id,
                    cluster_handler_unique_id=self.unique_id,
                    success=False,
                ),
            )

    async def configure_reporting(self) -> None:
        """Configure attribute reporting for a cluster.

        This also swallows ZigbeeException exceptions that are thrown when
        devices are unreachable.
        """
        event_data = {}
        kwargs = {}
        if (
            self.cluster.cluster_id >= 0xFC00
            and self._endpoint.device.manufacturer_code
        ):
            kwargs["manufacturer"] = self._endpoint.device.manufacturer_code

        for attr_report in self.REPORT_CONFIG:
            attr, config = attr_report["attr"], attr_report["config"]

            try:
                attr_name = self.cluster.find_attribute(attr).name
            except KeyError:
                attr_name = attr

            event_data[attr_name] = {
                "min": config[0],
                "max": config[1],
                "id": attr,
                "name": attr_name,
                "change": config[2],
                "status": None,
            }

        to_configure = [*self.REPORT_CONFIG]
        chunk, rest = (
            to_configure[:REPORT_CONFIG_ATTR_PER_REQ],
            to_configure[REPORT_CONFIG_ATTR_PER_REQ:],
        )
        while chunk:
            reports = {rec["attr"]: rec["config"] for rec in chunk}
            try:
                res = await RETRYABLE_REQUEST_DECORATOR(
                    self.cluster.configure_reporting_multiple
                )(reports, **kwargs)
                self._configure_reporting_status(reports, res[0], event_data)
            except (zigpy.exceptions.ZigbeeException, TimeoutError) as ex:
                self.debug(
                    "failed to set reporting on '%s' cluster for: %s",
                    self.cluster.ep_attribute,
                    str(ex),
                )
                break
            chunk, rest = (
                rest[:REPORT_CONFIG_ATTR_PER_REQ],
                rest[REPORT_CONFIG_ATTR_PER_REQ:],
            )

        self._endpoint.device.emit(
            ZHA_CLUSTER_HANDLER_MSG_CFG_RPT,
            ClusterConfigureReportingEvent(
                cluster_name=self.cluster.name,
                cluster_id=self.cluster.cluster_id,
                cluster_handler_unique_id=self.unique_id,
                attributes=event_data,
            ),
        )

    def _configure_reporting_status(
        self,
        attrs: dict[str, tuple[int, int, float | int]],
        res: list | tuple,
        event_data: dict[str, dict[str, Any]],
    ) -> None:
        """Parse configure reporting result."""
        if isinstance(res, (Exception, ConfigureReportingResponseRecord)):
            # assume default response
            self.debug(
                "attr reporting for '%s' on '%s': %s",
                attrs,
                self.name,
                res,
            )
            for attr in attrs:
                event_data[attr]["status"] = Status.FAILURE.name
            return
        if res[0].status == Status.SUCCESS and len(res) == 1:
            self.debug(
                "Successfully configured reporting for '%s' on '%s' cluster: %s",
                attrs,
                self.name,
                res,
            )
            # 2.5.8.1.3 Status Field
            # The status field specifies the status of the Configure Reporting operation attempted on this attribute,
            # as detailed in 2.5.7.3. Note that attribute status records are not included for successfully configured
            # attributes, in order to save bandwidth. In the case of successful configuration of all attributes,
            # only a single attribute status record SHALL be included in the command, with the status field set to
            # SUCCESS and the direction and attribute identifier fields omitted.
            for attr in attrs:
                event_data[attr]["status"] = Status.SUCCESS.name
            return

        for record in res:
            event_data[self.cluster.find_attribute(record.attrid).name]["status"] = (
                record.status.name
            )
        failed = [
            self.cluster.find_attribute(record.attrid).name
            for record in res
            if record.status != Status.SUCCESS
        ]
        self.debug(
            "Failed to configure reporting for '%s' on '%s' cluster: %s",
            failed,
            self.name,
            res,
        )
        success = set(attrs) - set(failed)
        self.debug(
            "Successfully configured reporting for '%s' on '%s' cluster",
            set(attrs) - set(failed),
            self.name,
        )
        for attr in success:
            event_data[attr]["status"] = Status.SUCCESS.name

    async def async_configure(self) -> None:
        """Set cluster binding and attribute reporting."""
        if self._endpoint.device.skip_configuration:
            self.debug("skipping cluster handler configuration")
            self._status = ClusterHandlerStatus.CONFIGURED
            return

        if self.BIND:
            self.debug("Performing cluster binding")
            await self.bind()

        if self.cluster.is_server:
            self.debug("Configuring cluster attribute reporting")
            await self.configure_reporting()

        self.debug("finished cluster handler configuration")
        self._status = ClusterHandlerStatus.CONFIGURED

    async def async_initialize(self, from_cache: bool) -> None:
        """Initialize cluster handler."""
        if not from_cache and self._endpoint.device.skip_configuration:
            self.debug("Skipping cluster handler initialization")
            self._status = ClusterHandlerStatus.INITIALIZED
            return

        self.debug("initializing cluster handler: from_cache: %s", from_cache)
        cached = [a for a, cached in self.ZCL_INIT_ATTRS.items() if cached]
        uncached = [a for a, cached in self.ZCL_INIT_ATTRS.items() if not cached]
        uncached.extend([cfg["attr"] for cfg in self.REPORT_CONFIG])

        if cached:
            self.debug("initializing cached cluster handler attributes: %s", cached)
            await self._get_attributes(
                True, cached, from_cache=True, only_cache=from_cache
            )
        if uncached:
            self.debug(
                "initializing uncached cluster handler attributes: %s - from cache[%s]",
                uncached,
                from_cache,
            )
            await self._get_attributes(
                True, uncached, from_cache=from_cache, only_cache=from_cache
            )

        ch_specific_init = getattr(
            self, "async_initialize_cluster_handler_specific", None
        )
        if ch_specific_init:
            self.debug(
                "Performing cluster handler specific initialization: %s", uncached
            )
            await ch_specific_init(from_cache=from_cache)

        self.debug("finished cluster handler initialization")
        self._status = ClusterHandlerStatus.INITIALIZED

    def cluster_command(self, tsn, command_id, args) -> None:
        """Handle commands received to this cluster."""

    def zdo_command(self, *args, **kwargs) -> None:
        """Handle ZDO commands on this cluster."""

    def zha_send_event(self, command: str, arg: list | dict | CommandSchema) -> None:
        """Compatibility method for emitting ZHA events from quirks until they are updated."""
        self.emit_zha_event(command, arg)

    def emit_zha_event(self, command: str, arg: list | dict | CommandSchema) -> None:
        """Relay events to listeners."""

        args: list | dict
        if isinstance(arg, CommandSchema):
            args = [a for a in arg if a is not None]
            params = arg.as_dict()
        elif isinstance(arg, (list, dict)):
            # Quirks can directly send lists and dicts to ZHA this way
            args = arg
            params = {}
        else:
            raise TypeError(f"Unexpected emit_zha_event {command!r} argument: {arg!r}")

        self._endpoint.emit_zha_event(
            {
                UNIQUE_ID: self.unique_id,
                CLUSTER_ID: self.cluster.cluster_id,
                COMMAND: command,
                # Maintain backwards compatibility with the old zigpy response format
                ARGS: args,
                PARAMS: params,
            }
        )

    async def async_update(self) -> None:
        """Retrieve latest state from cluster."""

    def _get_attribute_name(self, attrid: int) -> str | int:
        if attrid not in self.cluster.attributes:
            return attrid

        return self.cluster.attributes[attrid].name

    async def get_attribute_value(self, attribute, from_cache=True) -> Any:
        """Get the value for an attribute."""
        result = await safe_read(
            self._cluster,
            [attribute],
            allow_cache=from_cache,
            only_cache=from_cache,
        )
        return result.get(attribute)

    async def _get_attributes(
        self,
        raise_exceptions: bool,
        attributes: list[str],
        from_cache: bool = True,
        only_cache: bool = True,
    ) -> dict[int | str, Any]:
        """Get the values for a list of attributes."""
        chunk = attributes[:CLUSTER_READS_PER_REQ]
        rest = attributes[CLUSTER_READS_PER_REQ:]
        result = {}
        while chunk:
            try:
                self.debug("Reading attributes in chunks: %s", chunk)
                read, _ = await RETRYABLE_REQUEST_DECORATOR(
                    self.cluster.read_attributes
                )(
                    chunk,
                    allow_cache=from_cache,
                    only_cache=only_cache,
                    manufacturer=UNDEFINED,  # some quirks override default with None
                )
                self.debug("Got attributes: %s", read)
                result.update(read)
            except (TimeoutError, zigpy.exceptions.ZigbeeException) as ex:
                self.debug(
                    "failed to get attributes '%s' on '%s' cluster: %s",
                    chunk,
                    self.cluster.ep_attribute,
                    str(ex),
                )
                if raise_exceptions:
                    raise
            chunk = rest[:CLUSTER_READS_PER_REQ]
            rest = rest[CLUSTER_READS_PER_REQ:]
        return result

    async def get_attributes(
        self,
        attributes: list[str],
        from_cache: bool = True,
        only_cache: bool = True,
    ) -> dict[int | str, Any]:
        """Get the values for a list of attributes and raise no exceptions."""
        return await self._get_attributes(False, attributes, from_cache, only_cache)

    async def write_attributes_safe(
        self,
        attributes: dict[str, Any],
        manufacturer: int | UndefinedType | None = UNDEFINED,
    ) -> None:
        """Wrap `write_attributes` to throw an exception on attribute write failure."""

        res = await self.write_attributes(attributes, manufacturer=manufacturer)
        for record in res[0]:
            if record.status != Status.SUCCESS:
                try:
                    name = self.cluster.attributes[record.attrid].name
                    value = attributes.get(name, "unknown")
                except KeyError:
                    name = f"0x{record.attrid:04x}"
                    value = "unknown"

                raise ZHAException(
                    f"Failed to write attribute {name}={value}: {record.status}",
                )

    def log(self, level, msg, *args, **kwargs) -> None:
        """Log a message."""
        msg = f"[%s:%s]: {msg}"
        args = (self._endpoint.device.nwk, self._id) + args
        _LOGGER.log(level, msg, *args, stacklevel=3, **kwargs)

    def __getattr__(self, name):
        """Get attribute or a decorated cluster command."""
        if (
            hasattr(self._cluster, name)
            and callable(getattr(self._cluster, name))
            and name not in UNPROXIED_CLUSTER_METHODS
        ):
            command = getattr(self._cluster, name)
            wrapped_command = retry_request(command)
            wrapped_command.__name__ = name

            return wrapped_command
        return self.__getattribute__(name)


class ZDOClusterHandler(LogMixin):
    """Cluster handler for ZDO events."""

    def __init__(self, device) -> None:
        """Initialize ZDOClusterHandler."""
        self.name = CLUSTER_HANDLER_ZDO
        self._cluster = device.device.endpoints[0]
        self._zha_device = device
        self._status = ClusterHandlerStatus.CREATED
        self._unique_id = f"{str(device.ieee)}:{device.name}_ZDO"

    def on_add(self) -> None:
        """Call when cluster handler is added."""
        self._cluster.add_listener(self)

    def on_remove(self) -> None:
        """Call when cluster handler will be removed."""
        self._cluster.remove_listener(self)

    @property
    def unique_id(self):
        """Return the unique id for this cluster handler."""
        return self._unique_id

    @property
    def cluster(self):
        """Return the aigpy cluster for this cluster handler."""
        return self._cluster

    @property
    def status(self):
        """Return the status of the cluster handler."""
        return self._status

    def device_announce(self, zigpy_device):
        """Device announce handler."""

    def permit_duration(self, duration):
        """Permit handler."""

    async def async_initialize(self, from_cache):  # pylint: disable=unused-argument
        """Initialize cluster handler."""
        self._status = ClusterHandlerStatus.INITIALIZED

    async def async_configure(self):
        """Configure cluster handler."""
        self._status = ClusterHandlerStatus.CONFIGURED

    def log(self, level, msg, *args, **kwargs):
        """Log a message."""
        msg = f"[%s:ZDO](%s): {msg}"
        args = (self._zha_device.nwk, self._zha_device.model) + args
        _LOGGER.log(level, msg, *args, **kwargs)


class ClientClusterHandler(ClusterHandler):
    """ClusterHandler for Zigbee client (output) clusters."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize ClientClusterHandler."""
        super().__init__(*args, **kwargs)
        self._unique_id += "_CLIENT"
        self._generic_id += "_client"
        self._id += "_client"

    def _handle_attribute_updated_event(
        self,
        event: AttributeReadEvent
        | AttributeReportedEvent
        | AttributeUpdatedEvent
        | AttributeWrittenEvent,
    ) -> None:
        """Handle an attribute updated on this cluster."""
        super()._handle_attribute_updated_event(event)

        self.emit_zha_event(
            SIGNAL_ATTR_UPDATED,
            {
                ATTRIBUTE_ID: event.attribute_id,
                ATTRIBUTE_NAME: event.attribute_name or "Unknown",
                ATTRIBUTE_VALUE: event.value,
                VALUE: event.value,
            },
        )

    def cluster_command(self, tsn, command_id, args):
        """Handle a cluster command received on this cluster."""
        if (
            self._cluster.server_commands is not None
            and self._cluster.server_commands.get(command_id) is not None
        ):
            self.emit_zha_event(self._cluster.server_commands[command_id].name, args)



"""BACnet units."""

import zigpy.types as t

from zha.units import (
    CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    CONCENTRATION_MILLIGRAMS_PER_CUBIC_METER,
    CONCENTRATION_PARTS_PER_BILLION,
    CONCENTRATION_PARTS_PER_MILLION,
    DEGREE,
    LIGHT_LUX,
    PERCENTAGE,
    REVOLUTIONS_PER_MINUTE,
    SIGNAL_STRENGTH_DECIBELS,
    SIGNAL_STRENGTH_DECIBELS_MILLIWATT,
    UnitOfApparentPower,
    UnitOfArea,
    UnitOfConductivity,
    UnitOfElectricCurrent,
    UnitOfElectricPotential,
    UnitOfEnergy,
    UnitOfEnergyDistance,
    UnitOfFrequency,
    UnitOfIrradiance,
    UnitOfLength,
    UnitOfMass,
    UnitOfPower,
    UnitOfPressure,
    UnitOfReactivePower,
    UnitOfSoundPressure,
    UnitOfSpeed,
    UnitOfTemperature,
    UnitOfTime,
    UnitOfVolume,
    UnitOfVolumeFlowRate,
)


# https://github.com/bacnet-stack/bacnet-stack/blob/980c78dddbd849c7ff8436f317058c7905213f6d/src/bacnet/bacenum.h
class BacnetUnit(t.enum8):
    """BACnet units."""

    SQUARE_METERS = 0
    SQUARE_FEET = 1
    MILLIAMPERES = 2
    AMPERES = 3
    OHMS = 4
    VOLTS = 5
    KILOVOLTS = 6
    MEGAVOLTS = 7
    VOLT_AMPERES = 8
    KILOVOLT_AMPERES = 9
    MEGAVOLT_AMPERES = 10
    VOLT_AMPERES_REACTIVE = 11
    KILOVOLT_AMPERES_REACTIVE = 12
    MEGAVOLT_AMPERES_REACTIVE = 13
    DEGREES_PHASE = 14
    POWER_FACTOR = 15
    JOULES = 16
    KILOJOULES = 17
    WATT_HOURS = 18
    KILOWATT_HOURS = 19
    BTUS = 20
    THERMS = 21
    TON_HOURS = 22
    JOULES_PER_KILOGRAM_DRY_AIR = 23
    BTUS_PER_POUND_DRY_AIR = 24
    CYCLES_PER_HOUR = 25
    CYCLES_PER_MINUTE = 26
    HERTZ = 27
    GRAMS_OF_WATER_PER_KILOGRAM_DRY_AIR = 28
    PERCENT_RELATIVE_HUMIDITY = 29
    MILLIMETERS = 30
    METERS = 31
    INCHES = 32
    FEET = 33
    WATTS_PER_SQUARE_FOOT = 34
    WATTS_PER_SQUARE_METER = 35
    LUMENS = 36
    LUXES = 37
    FOOT_CANDLES = 38
    KILOGRAMS = 39
    POUNDS_MASS = 40
    TONS = 41
    KILOGRAMS_PER_SECOND = 42
    KILOGRAMS_PER_MINUTE = 43
    KILOGRAMS_PER_HOUR = 44
    POUNDS_MASS_PER_MINUTE = 45
    POUNDS_MASS_PER_HOUR = 46
    WATTS = 47
    KILOWATTS = 48
    MEGAWATTS = 49
    BTUS_PER_HOUR = 50
    HORSEPOWER = 51
    TONS_REFRIGERATION = 52
    PASCALS = 53
    KILOPASCALS = 54
    BARS = 55
    POUNDS_FORCE_PER_SQUARE_INCH = 56
    CENTIMETERS_OF_WATER = 57
    INCHES_OF_WATER = 58
    MILLIMETERS_OF_MERCURY = 59
    CENTIMETERS_OF_MERCURY = 60
    INCHES_OF_MERCURY = 61
    DEGREES_CELSIUS = 62
    KELVIN = 63
    DEGREES_FAHRENHEIT = 64
    DEGREE_DAYS_CELSIUS = 65
    DEGREE_DAYS_FAHRENHEIT = 66
    YEARS = 67
    MONTHS = 68
    WEEKS = 69
    DAYS = 70
    HOURS = 71
    MINUTES = 72
    SECONDS = 73
    METERS_PER_SECOND = 74
    KILOMETERS_PER_HOUR = 75
    FEET_PER_SECOND = 76
    FEET_PER_MINUTE = 77
    MILES_PER_HOUR = 78
    CUBIC_FEET = 79
    CUBIC_METERS = 80
    IMPERIAL_GALLONS = 81
    LITERS = 82
    US_GALLONS = 83
    CUBIC_FEET_PER_MINUTE = 84
    CUBIC_METERS_PER_SECOND = 85
    IMPERIAL_GALLONS_PER_MINUTE = 86
    LITERS_PER_SECOND = 87
    LITERS_PER_MINUTE = 88
    US_GALLONS_PER_MINUTE = 89
    DEGREES_ANGULAR = 90
    DEGREES_CELSIUS_PER_HOUR = 91
    DEGREES_CELSIUS_PER_MINUTE = 92
    DEGREES_FAHRENHEIT_PER_HOUR = 93
    DEGREES_FAHRENHEIT_PER_MINUTE = 94
    NO_UNITS = 95
    PARTS_PER_MILLION = 96
    PARTS_PER_BILLION = 97
    PERCENT = 98
    PERCENT_PER_SECOND = 99
    PER_MINUTE = 100
    PER_SECOND = 101
    PSI_PER_DEGREE_FAHRENHEIT = 102
    RADIANS = 103
    REVOLUTIONS_PER_MINUTE = 104
    CURRENCY1 = 105
    CURRENCY2 = 106
    CURRENCY3 = 107
    CURRENCY4 = 108
    CURRENCY5 = 109
    CURRENCY6 = 110
    CURRENCY7 = 111
    CURRENCY8 = 112
    CURRENCY9 = 113
    CURRENCY10 = 114
    SQUARE_INCHES = 115
    SQUARE_CENTIMETERS = 116
    BTUS_PER_POUND = 117
    CENTIMETERS = 118
    POUNDS_MASS_PER_SECOND = 119
    DELTA_DEGREES_FAHRENHEIT = 120
    DELTA_KELVIN = 121
    KILOHMS = 122
    MEGOHMS = 123
    MILLIVOLTS = 124
    KILOJOULES_PER_KILOGRAM = 125
    MEGAJOULES = 126
    JOULES_PER_DEGREE_KELVIN = 127
    JOULES_PER_KILOGRAM_DEGREE_KELVIN = 128
    KILOHERTZ = 129
    MEGAHERTZ = 130
    PER_HOUR = 131
    MILLIWATTS = 132
    HECTOPASCALS = 133
    MILLIBARS = 134
    CUBIC_METERS_PER_HOUR = 135
    LITERS_PER_HOUR = 136
    KW_HOURS_PER_SQUARE_METER = 137
    KW_HOURS_PER_SQUARE_FOOT = 138
    MEGAJOULES_PER_SQUARE_METER = 139
    MEGAJOULES_PER_SQUARE_FOOT = 140
    WATTS_PER_SQUARE_METER_DEGREE_KELVIN = 141
    CUBIC_FEET_PER_SECOND = 142
    PERCENT_OBSCURATION_PER_FOOT = 143
    PERCENT_OBSCURATION_PER_METER = 144
    MILLIOHMS = 145
    MEGAWATT_HOURS = 146
    KILO_BTUS = 147
    MEGA_BTUS = 148
    KILOJOULES_PER_KILOGRAM_DRY_AIR = 149
    MEGAJOULES_PER_KILOGRAM_DRY_AIR = 150
    KILOJOULES_PER_DEGREE_KELVIN = 151
    MEGAJOULES_PER_DEGREE_KELVIN = 152
    NEWTON = 153
    GRAMS_PER_SECOND = 154
    GRAMS_PER_MINUTE = 155
    TONS_PER_HOUR = 156
    KILO_BTUS_PER_HOUR = 157
    HUNDREDTHS_SECONDS = 158
    MILLISECONDS = 159
    NEWTON_METERS = 160
    MILLIMETERS_PER_SECOND = 161
    MILLIMETERS_PER_MINUTE = 162
    METERS_PER_MINUTE = 163
    METERS_PER_HOUR = 164
    CUBIC_METERS_PER_MINUTE = 165
    METERS_PER_SECOND_PER_SECOND = 166
    AMPERES_PER_METER = 167
    AMPERES_PER_SQUARE_METER = 168
    AMPERE_SQUARE_METERS = 169
    FARADS = 170
    HENRYS = 171
    OHM_METERS = 172
    SIEMENS_PER_METER = 174
    TESLAS = 175
    VOLTS_PER_DEGREE_KELVIN = 176
    VOLTS_PER_METER = 177
    WEBERS = 178
    CANDELAS = 179
    CANDELAS_PER_SQUARE_METER = 180
    KELVIN_PER_HOUR = 181
    KELVIN_PER_MINUTE = 182
    JOULE_SECONDS = 183
    RADIANS_PER_SECOND = 184
    SQUARE_METERS_PER_NEWTON = 185
    KILOGRAMS_PER_CUBIC_METER = 186
    NEWTON_SECONDS = 187
    NEWTONS_PER_METER = 188
    WATTS_PER_METER_PER_DEGREE_KELVIN = 189
    MICROSIEMENS = 190
    CUBIC_FEET_PER_HOUR = 191
    US_GALLONS_PER_HOUR = 192
    KILOMETERS = 193
    MICROMETERS = 194
    GRAMS = 195
    MILLIGRAMS = 196
    MILLILITERS = 197
    MILLILITERS_PER_SECOND = 198
    DECIBELS = 199
    DECIBELS_MILLIVOLT = 200
    DECIBELS_VOLT = 201
    MILLISIEMENS = 202
    WATT_REACTIVE_HOURS = 203
    KILOWATT_REACTIVE_HOURS = 204
    MEGAWATT_REACTIVE_HOURS = 205
    MILLIMETERS_OF_WATER = 206
    PER_MILLE = 207
    GRAMS_PER_GRAM = 208
    KILOGRAMS_PER_KILOGRAM = 209
    GRAMS_PER_KILOGRAM = 210
    MILLIGRAMS_PER_GRAM = 211
    MILLIGRAMS_PER_KILOGRAM = 212
    GRAMS_PER_MILLILITER = 213
    GRAMS_PER_LITER = 214
    MILLIGRAMS_PER_LITER = 215
    MICROGRAMS_PER_LITER = 216
    GRAMS_PER_CUBIC_METER = 217
    MILLIGRAMS_PER_CUBIC_METER = 218
    MICROGRAMS_PER_CUBIC_METER = 219
    NANOGRAMS_PER_CUBIC_METER = 220
    GRAMS_PER_CUBIC_CENTIMETER = 221
    BECQUERELS = 222
    KILOBECQUERELS = 223
    MEGABECQUERELS = 224
    GRAY = 225
    MILLIGRAY = 226
    MICROGRAY = 227
    SIEVERTS = 228
    MILLISIEVERTS = 229
    MICROSIEVERTS = 230
    MICROSIEVERTS_PER_HOUR = 231
    DECIBELS_A = 232
    NEPHELOMETRIC_TURBIDITY_UNIT = 233
    PH = 234
    GRAMS_PER_SQUARE_METER = 235
    MINUTES_PER_DEGREE_KELVIN = 236
    OHM_METER_SQUARED_PER_METER = 237
    AMPERE_SECONDS = 238
    VOLT_AMPERE_HOURS = 239
    KILOVOLT_AMPERE_HOURS = 240
    MEGAVOLT_AMPERE_HOURS = 241
    VOLT_AMPERE_REACTIVE_HOURS = 242
    KILOVOLT_AMPERE_REACTIVE_HOURS = 243
    MEGAVOLT_AMPERE_REACTIVE_HOURS = 244
    VOLT_SQUARE_HOURS = 245
    AMPERE_SQUARE_HOURS = 246
    JOULE_PER_HOURS = 247
    CUBIC_FEET_PER_DAY = 248
    CUBIC_METERS_PER_DAY = 249
    WATT_HOURS_PER_CUBIC_METER = 250
    JOULES_PER_CUBIC_METER = 251
    MOLE_PERCENT = 252
    PASCAL_SECONDS = 253
    MILLION_STANDARD_CUBIC_FEET_PER_MINUTE = 254


BACNET_UNITS_TO_HA_UNITS = {
    BacnetUnit.SQUARE_METERS: UnitOfArea.SQUARE_METERS,
    BacnetUnit.SQUARE_FEET: UnitOfArea.SQUARE_FEET,
    BacnetUnit.MILLIAMPERES: UnitOfElectricCurrent.MILLIAMPERE,
    BacnetUnit.AMPERES: UnitOfElectricCurrent.AMPERE,
    BacnetUnit.OHMS: "Ω",
    BacnetUnit.VOLTS: UnitOfElectricPotential.VOLT,
    BacnetUnit.KILOVOLTS: UnitOfElectricPotential.KILOVOLT,
    BacnetUnit.MEGAVOLTS: UnitOfElectricPotential.MEGAVOLT,
    BacnetUnit.VOLT_AMPERES: UnitOfApparentPower.VOLT_AMPERE,
    BacnetUnit.KILOVOLT_AMPERES: "kVA",
    BacnetUnit.MEGAVOLT_AMPERES: "MVA",
    BacnetUnit.VOLT_AMPERES_REACTIVE: UnitOfReactivePower.VOLT_AMPERE_REACTIVE,
    BacnetUnit.KILOVOLT_AMPERES_REACTIVE: UnitOfReactivePower.KILO_VOLT_AMPERE_REACTIVE,
    BacnetUnit.MEGAVOLT_AMPERES_REACTIVE: "Mvar",
    BacnetUnit.DEGREES_PHASE: DEGREE,
    BacnetUnit.POWER_FACTOR: None,
    BacnetUnit.JOULES: UnitOfEnergy.JOULE,
    BacnetUnit.KILOJOULES: UnitOfEnergy.KILO_JOULE,
    BacnetUnit.WATT_HOURS: UnitOfEnergy.WATT_HOUR,
    BacnetUnit.KILOWATT_HOURS: UnitOfEnergy.KILO_WATT_HOUR,
    BacnetUnit.BTUS: UnitOfPower.BTU_PER_HOUR,
    BacnetUnit.THERMS: "therm",
    BacnetUnit.TON_HOURS: "ton·h",
    BacnetUnit.JOULES_PER_KILOGRAM_DRY_AIR: "J/kg",
    BacnetUnit.BTUS_PER_POUND_DRY_AIR: "BTU/lb",
    BacnetUnit.CYCLES_PER_HOUR: "cycles/h",
    BacnetUnit.CYCLES_PER_MINUTE: "cycles/m",
    BacnetUnit.HERTZ: UnitOfFrequency.HERTZ,
    BacnetUnit.GRAMS_OF_WATER_PER_KILOGRAM_DRY_AIR: "g/kg",
    BacnetUnit.PERCENT_RELATIVE_HUMIDITY: PERCENTAGE,
    BacnetUnit.MILLIMETERS: UnitOfLength.MILLIMETERS,
    BacnetUnit.METERS: UnitOfLength.METERS,
    BacnetUnit.INCHES: UnitOfLength.INCHES,
    BacnetUnit.FEET: UnitOfLength.FEET,
    BacnetUnit.WATTS_PER_SQUARE_FOOT: "W/ft²",
    BacnetUnit.WATTS_PER_SQUARE_METER: UnitOfIrradiance.WATTS_PER_SQUARE_METER,
    BacnetUnit.LUMENS: "lm",
    BacnetUnit.LUXES: LIGHT_LUX,
    BacnetUnit.FOOT_CANDLES: "fc",
    BacnetUnit.KILOGRAMS: UnitOfMass.KILOGRAMS,
    BacnetUnit.POUNDS_MASS: UnitOfMass.POUNDS,
    BacnetUnit.TONS: UnitOfMass.STONES,
    BacnetUnit.KILOGRAMS_PER_SECOND: "kg/s",
    BacnetUnit.KILOGRAMS_PER_MINUTE: "kg/min",
    BacnetUnit.KILOGRAMS_PER_HOUR: "kg/h",
    BacnetUnit.POUNDS_MASS_PER_MINUTE: "lb/min",
    BacnetUnit.POUNDS_MASS_PER_HOUR: "lb/h",
    BacnetUnit.WATTS: UnitOfPower.WATT,
    BacnetUnit.KILOWATTS: UnitOfPower.KILO_WATT,
    BacnetUnit.MEGAWATTS: UnitOfPower.MEGA_WATT,
    BacnetUnit.BTUS_PER_HOUR: UnitOfPower.BTU_PER_HOUR,
    BacnetUnit.HORSEPOWER: "hp",
    BacnetUnit.TONS_REFRIGERATION: "ton",
    BacnetUnit.PASCALS: UnitOfPressure.PA,
    BacnetUnit.KILOPASCALS: UnitOfPressure.KPA,
    BacnetUnit.BARS: UnitOfPressure.BAR,
    BacnetUnit.POUNDS_FORCE_PER_SQUARE_INCH: UnitOfPressure.PSI,
    BacnetUnit.CENTIMETERS_OF_WATER: "cm H₂O",
    BacnetUnit.INCHES_OF_WATER: "in H₂O",
    BacnetUnit.MILLIMETERS_OF_MERCURY: UnitOfPressure.MMHG,
    BacnetUnit.CENTIMETERS_OF_MERCURY: "cm Hg",
    BacnetUnit.INCHES_OF_MERCURY: UnitOfPressure.INHG,
    BacnetUnit.DEGREES_CELSIUS: UnitOfTemperature.CELSIUS,
    BacnetUnit.KELVIN: UnitOfTemperature.KELVIN,
    BacnetUnit.DEGREES_FAHRENHEIT: UnitOfTemperature.FAHRENHEIT,
    BacnetUnit.DEGREE_DAYS_CELSIUS: "°C·d",
    BacnetUnit.DEGREE_DAYS_FAHRENHEIT: "°F·d",
    BacnetUnit.YEARS: UnitOfTime.YEARS,
    BacnetUnit.MONTHS: UnitOfTime.MONTHS,
    BacnetUnit.WEEKS: UnitOfTime.WEEKS,
    BacnetUnit.DAYS: UnitOfTime.DAYS,
    BacnetUnit.HOURS: UnitOfTime.HOURS,
    BacnetUnit.MINUTES: UnitOfTime.MINUTES,
    BacnetUnit.SECONDS: UnitOfTime.SECONDS,
    BacnetUnit.METERS_PER_SECOND: UnitOfSpeed.METERS_PER_SECOND,
    BacnetUnit.KILOMETERS_PER_HOUR: UnitOfSpeed.KILOMETERS_PER_HOUR,
    BacnetUnit.FEET_PER_SECOND: UnitOfSpeed.FEET_PER_SECOND,
    BacnetUnit.FEET_PER_MINUTE: "ft/min",
    BacnetUnit.MILES_PER_HOUR: UnitOfSpeed.MILES_PER_HOUR,
    BacnetUnit.CUBIC_FEET: UnitOfVolume.CUBIC_FEET,
    BacnetUnit.CUBIC_METERS: UnitOfVolume.CUBIC_METERS,
    BacnetUnit.IMPERIAL_GALLONS: "imperial gal",
    BacnetUnit.LITERS: UnitOfVolume.LITERS,
    BacnetUnit.US_GALLONS: UnitOfVolume.GALLONS,
    BacnetUnit.CUBIC_FEET_PER_MINUTE: UnitOfVolumeFlowRate.CUBIC_FEET_PER_MINUTE,
    BacnetUnit.CUBIC_METERS_PER_SECOND: "m³/s",
    BacnetUnit.IMPERIAL_GALLONS_PER_MINUTE: "imperial gal/min",
    BacnetUnit.LITERS_PER_SECOND: "L/s",
    BacnetUnit.LITERS_PER_MINUTE: UnitOfVolumeFlowRate.LITERS_PER_MINUTE,
    BacnetUnit.US_GALLONS_PER_MINUTE: UnitOfVolumeFlowRate.GALLONS_PER_MINUTE,
    BacnetUnit.DEGREES_ANGULAR: REVOLUTIONS_PER_MINUTE,
    BacnetUnit.DEGREES_CELSIUS_PER_HOUR: "°C/h",
    BacnetUnit.DEGREES_CELSIUS_PER_MINUTE: "°C/min",
    BacnetUnit.DEGREES_FAHRENHEIT_PER_HOUR: "°F/h",
    BacnetUnit.DEGREES_FAHRENHEIT_PER_MINUTE: "°F/min",
    BacnetUnit.NO_UNITS: None,
    BacnetUnit.PARTS_PER_MILLION: CONCENTRATION_PARTS_PER_MILLION,
    BacnetUnit.PARTS_PER_BILLION: CONCENTRATION_PARTS_PER_BILLION,
    BacnetUnit.PERCENT: PERCENTAGE,
    BacnetUnit.PERCENT_PER_SECOND: "%/s",
    BacnetUnit.PER_MINUTE: "1/min",
    BacnetUnit.PER_SECOND: "1/s",
    BacnetUnit.PSI_PER_DEGREE_FAHRENHEIT: "psi/°F",
    BacnetUnit.RADIANS: "rad",
    BacnetUnit.REVOLUTIONS_PER_MINUTE: REVOLUTIONS_PER_MINUTE,
    BacnetUnit.CURRENCY1: "currency",
    BacnetUnit.CURRENCY2: "currency",
    BacnetUnit.CURRENCY3: "currency",
    BacnetUnit.CURRENCY4: "currency",
    BacnetUnit.CURRENCY5: "currency",
    BacnetUnit.CURRENCY6: "currency",
    BacnetUnit.CURRENCY7: "currency",
    BacnetUnit.CURRENCY8: "currency",
    BacnetUnit.CURRENCY9: "currency",
    BacnetUnit.CURRENCY10: "currency",
    BacnetUnit.SQUARE_INCHES: UnitOfArea.SQUARE_INCHES,
    BacnetUnit.SQUARE_CENTIMETERS: UnitOfArea.SQUARE_CENTIMETERS,
    BacnetUnit.BTUS_PER_POUND: UnitOfEnergy.CALORIE,
    BacnetUnit.CENTIMETERS: UnitOfLength.CENTIMETERS,
    BacnetUnit.POUNDS_MASS_PER_SECOND: "lb/s",
    BacnetUnit.DELTA_DEGREES_FAHRENHEIT: "Δ°F",
    BacnetUnit.DELTA_KELVIN: "ΔK",
    BacnetUnit.KILOHMS: "kΩ",
    BacnetUnit.MEGOHMS: "MΩ",
    BacnetUnit.MILLIVOLTS: UnitOfElectricPotential.MILLIVOLT,
    BacnetUnit.KILOJOULES_PER_KILOGRAM: "kJ/kg",
    BacnetUnit.MEGAJOULES: UnitOfEnergy.MEGA_JOULE,
    BacnetUnit.JOULES_PER_DEGREE_KELVIN: "J/K",
    BacnetUnit.JOULES_PER_KILOGRAM_DEGREE_KELVIN: "J/(kg·K)",
    BacnetUnit.KILOHERTZ: UnitOfFrequency.KILOHERTZ,
    BacnetUnit.MEGAHERTZ: UnitOfFrequency.MEGAHERTZ,
    BacnetUnit.PER_HOUR: "1/h",
    BacnetUnit.MILLIWATTS: UnitOfPower.MILLIWATT,
    BacnetUnit.HECTOPASCALS: UnitOfPressure.HPA,
    BacnetUnit.MILLIBARS: UnitOfPressure.MBAR,
    BacnetUnit.CUBIC_METERS_PER_HOUR: UnitOfVolumeFlowRate.CUBIC_METERS_PER_HOUR,
    BacnetUnit.LITERS_PER_HOUR: UnitOfVolumeFlowRate.LITERS_PER_MINUTE,
    BacnetUnit.KW_HOURS_PER_SQUARE_METER: UnitOfEnergyDistance.KM_PER_KILO_WATT_HOUR,
    BacnetUnit.KW_HOURS_PER_SQUARE_FOOT: UnitOfEnergyDistance.MILES_PER_KILO_WATT_HOUR,
    BacnetUnit.MEGAJOULES_PER_SQUARE_METER: f"{UnitOfEnergy.MEGA_JOULE}/m²",
    BacnetUnit.MEGAJOULES_PER_SQUARE_FOOT: f"{UnitOfEnergy.MEGA_JOULE}/ft²",
    BacnetUnit.WATTS_PER_SQUARE_METER_DEGREE_KELVIN: "W/(m²·K)",
    BacnetUnit.CUBIC_FEET_PER_SECOND: "ft³/s",
    BacnetUnit.PERCENT_OBSCURATION_PER_FOOT: "% obscuration/ft",
    BacnetUnit.PERCENT_OBSCURATION_PER_METER: "% obscuration/m",
    BacnetUnit.MILLIOHMS: "mΩ",
    BacnetUnit.MEGAWATT_HOURS: UnitOfEnergy.MEGA_WATT_HOUR,
    BacnetUnit.KILO_BTUS: "kBTU",
    BacnetUnit.MEGA_BTUS: "MBTU",
    BacnetUnit.KILOJOULES_PER_KILOGRAM_DRY_AIR: "kJ/kg_dry",
    BacnetUnit.MEGAJOULES_PER_KILOGRAM_DRY_AIR: "MJ/kg_dry",
    BacnetUnit.KILOJOULES_PER_DEGREE_KELVIN: "kJ/K",
    BacnetUnit.MEGAJOULES_PER_DEGREE_KELVIN: "MJ/K",
    BacnetUnit.NEWTON: "N",
    BacnetUnit.GRAMS_PER_SECOND: "g/s",
    BacnetUnit.GRAMS_PER_MINUTE: "g/min",
    BacnetUnit.TONS_PER_HOUR: "tons/h",
    BacnetUnit.KILO_BTUS_PER_HOUR: "kBTU/h",
    BacnetUnit.HUNDREDTHS_SECONDS: "1/100 s",
    BacnetUnit.MILLISECONDS: UnitOfTime.MILLISECONDS,
    BacnetUnit.NEWTON_METERS: "N·m",
    BacnetUnit.MILLIMETERS_PER_SECOND: UnitOfLength.MILLIMETERS,
    BacnetUnit.MILLIMETERS_PER_MINUTE: "mm/min",
    BacnetUnit.METERS_PER_MINUTE: "m/min",
    BacnetUnit.METERS_PER_HOUR: "m/h",
    BacnetUnit.CUBIC_METERS_PER_MINUTE: "m³/min",
    BacnetUnit.METERS_PER_SECOND_PER_SECOND: "m/s²",
    BacnetUnit.AMPERES_PER_METER: "A/m",
    BacnetUnit.AMPERES_PER_SQUARE_METER: "A/m²",
    BacnetUnit.AMPERE_SQUARE_METERS: "A·m²",
    BacnetUnit.FARADS: "F",
    BacnetUnit.HENRYS: "H",
    BacnetUnit.OHM_METERS: "Ω·m",
    BacnetUnit.SIEMENS_PER_METER: "S/m",
    BacnetUnit.TESLAS: "T",
    BacnetUnit.VOLTS_PER_DEGREE_KELVIN: "V/K",
    BacnetUnit.VOLTS_PER_METER: "V/m",
    BacnetUnit.WEBERS: "Wb",
    BacnetUnit.CANDELAS: "cd",
    BacnetUnit.CANDELAS_PER_SQUARE_METER: "cd/m²",
    BacnetUnit.KELVIN_PER_HOUR: "K/h",
    BacnetUnit.KELVIN_PER_MINUTE: "K/min",
    BacnetUnit.JOULE_SECONDS: "J·s",
    BacnetUnit.RADIANS_PER_SECOND: "rad/s",
    BacnetUnit.SQUARE_METERS_PER_NEWTON: "m²/N",
    BacnetUnit.KILOGRAMS_PER_CUBIC_METER: "kg/m³",
    BacnetUnit.NEWTON_SECONDS: "N·s",
    BacnetUnit.NEWTONS_PER_METER: "N/m",
    BacnetUnit.WATTS_PER_METER_PER_DEGREE_KELVIN: "W/(m·K)",
    BacnetUnit.MICROSIEMENS: UnitOfConductivity.MICROSIEMENS_PER_CM,
    BacnetUnit.CUBIC_FEET_PER_HOUR: "ft³/h",
    BacnetUnit.US_GALLONS_PER_HOUR: "gal/h",
    BacnetUnit.KILOMETERS: UnitOfLength.KILOMETERS,
    BacnetUnit.MICROMETERS: "µm",
    BacnetUnit.GRAMS: UnitOfMass.GRAMS,
    BacnetUnit.MILLIGRAMS: UnitOfMass.MILLIGRAMS,
    BacnetUnit.MILLILITERS: UnitOfVolume.MILLILITERS,
    BacnetUnit.MILLILITERS_PER_SECOND: UnitOfVolumeFlowRate.MILLILITERS_PER_SECOND,
    BacnetUnit.DECIBELS: UnitOfSoundPressure.DECIBEL,
    BacnetUnit.DECIBELS_MILLIVOLT: SIGNAL_STRENGTH_DECIBELS_MILLIWATT,
    BacnetUnit.DECIBELS_VOLT: SIGNAL_STRENGTH_DECIBELS,
    BacnetUnit.MILLISIEMENS: UnitOfConductivity.MILLISIEMENS_PER_CM,
    BacnetUnit.WATT_REACTIVE_HOURS: "var·h",
    BacnetUnit.KILOWATT_REACTIVE_HOURS: "kvar·h",
    BacnetUnit.MEGAWATT_REACTIVE_HOURS: "Mvar·h",
    BacnetUnit.MILLIMETERS_OF_WATER: "mm H₂O",
    BacnetUnit.PER_MILLE: "‰",
    BacnetUnit.GRAMS_PER_GRAM: "g/g",
    BacnetUnit.KILOGRAMS_PER_KILOGRAM: "kg/kg",
    BacnetUnit.GRAMS_PER_KILOGRAM: "g/kg",
    BacnetUnit.MILLIGRAMS_PER_GRAM: "mg/g",
    BacnetUnit.MILLIGRAMS_PER_KILOGRAM: "mg/kg",
    BacnetUnit.GRAMS_PER_MILLILITER: "g/mL",
    BacnetUnit.GRAMS_PER_LITER: "g/L",
    BacnetUnit.MILLIGRAMS_PER_LITER: "mg/L",
    BacnetUnit.MICROGRAMS_PER_LITER: "µg/L",
    BacnetUnit.GRAMS_PER_CUBIC_METER: "g/m³",
    BacnetUnit.MILLIGRAMS_PER_CUBIC_METER: CONCENTRATION_MILLIGRAMS_PER_CUBIC_METER,
    BacnetUnit.MICROGRAMS_PER_CUBIC_METER: CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    BacnetUnit.NANOGRAMS_PER_CUBIC_METER: "ng/m³",
    BacnetUnit.GRAMS_PER_CUBIC_CENTIMETER: "g/cm³",
    BacnetUnit.BECQUERELS: "Bq",
    BacnetUnit.KILOBECQUERELS: "kBq",
    BacnetUnit.MEGABECQUERELS: "MBq",
    BacnetUnit.GRAY: "Gy",
    BacnetUnit.MILLIGRAY: "mGy",
    BacnetUnit.MICROGRAY: "µGy",
    BacnetUnit.SIEVERTS: "Sv",
    BacnetUnit.MILLISIEVERTS: "mSv",
    BacnetUnit.MICROSIEVERTS: "µSv",
    BacnetUnit.MICROSIEVERTS_PER_HOUR: "µSv/h",
    BacnetUnit.DECIBELS_A: "dBA",
    BacnetUnit.NEPHELOMETRIC_TURBIDITY_UNIT: "NTU",
    BacnetUnit.PH: "pH",
    BacnetUnit.GRAMS_PER_SQUARE_METER: "g/m²",
    BacnetUnit.MINUTES_PER_DEGREE_KELVIN: "min/K",
    BacnetUnit.OHM_METER_SQUARED_PER_METER: "Ω·m²/m",
    BacnetUnit.AMPERE_SQUARE_HOURS: "A·h",
    BacnetUnit.VOLT_AMPERE_HOURS: "VA·h",
    BacnetUnit.KILOVOLT_AMPERE_HOURS: "kVA·h",
    BacnetUnit.MEGAVOLT_AMPERE_HOURS: "MVA·h",
    BacnetUnit.VOLT_AMPERE_REACTIVE_HOURS: "var·h",
    BacnetUnit.KILOVOLT_AMPERE_REACTIVE_HOURS: "kvar·h",
    BacnetUnit.MEGAVOLT_AMPERE_REACTIVE_HOURS: "Mvar·h",
    BacnetUnit.VOLT_SQUARE_HOURS: "V²·h",
    BacnetUnit.JOULE_PER_HOURS: "J/h",
    BacnetUnit.CUBIC_FEET_PER_DAY: "ft³/d",
    BacnetUnit.CUBIC_METERS_PER_DAY: "m³/d",
    BacnetUnit.WATT_HOURS_PER_CUBIC_METER: "Wh/m³",
    BacnetUnit.JOULES_PER_CUBIC_METER: "J/m³",
    BacnetUnit.MOLE_PERCENT: "mol%",
    BacnetUnit.PASCAL_SECONDS: "Pa·s",
    BacnetUnit.MILLION_STANDARD_CUBIC_FEET_PER_MINUTE: "MMSCFM",
}



"""Tests for the gateway module."""

import asyncio
import functools
import time
from unittest.mock import MagicMock, patch

import pytest

from zha import async_ as zha_async
from zha.application.gateway import Gateway
from zha.async_ import AsyncUtilMixin, ZHAJob, ZHAJobType, create_eager_task


@pytest.mark.parametrize("eager_start", [True, False])
async def test_cancellable_zhajob(zha_gateway: Gateway, eager_start: bool) -> None:
    """Simulate a shutdown, ensure cancellable jobs are cancelled."""
    job = MagicMock()

    def run_job(job: ZHAJob) -> None:
        """Call the action."""
        zha_gateway.async_run_zha_job(job, eager_start=eager_start)

    timer1 = zha_gateway.loop.call_later(
        60, run_job, ZHAJob(job, cancel_on_shutdown=True)
    )
    timer2 = zha_gateway.loop.call_later(60, run_job, ZHAJob(job))

    await zha_gateway.shutdown()

    assert timer1.cancelled()
    assert not timer2.cancelled()

    # Cleanup
    timer2.cancel()


async def test_async_add_zha_job_schedule_callback() -> None:
    """Test that we schedule callbacks and add jobs to the job pool."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))
    job = MagicMock()

    AsyncUtilMixin.async_add_zha_job(zha_gateway, ZHAJob(job))
    assert len(zha_gateway.loop.call_soon.mock_calls) == 1
    assert len(zha_gateway.loop.create_task.mock_calls) == 0
    assert len(zha_gateway.add_job.mock_calls) == 0


async def test_async_add_zha_job_eager_start_coro_suspends(
    zha_gateway: Gateway,
) -> None:
    """Test scheduling a coro as a task that will suspend with eager_start."""

    async def job_that_suspends():
        await asyncio.sleep(0)

    task = zha_gateway.async_add_zha_job(ZHAJob(job_that_suspends), eager_start=True)
    assert not task.done()
    assert task in zha_gateway._tracked_completable_tasks
    await task
    assert task not in zha_gateway._tracked_completable_tasks


async def test_async_run_zha_job_eager_start_coro_suspends(
    zha_gateway: Gateway,
) -> None:
    """Test scheduling a coro as a task that will suspend with eager_start."""

    async def job_that_suspends():
        await asyncio.sleep(0)

    task = zha_gateway.async_run_zha_job(ZHAJob(job_that_suspends))
    assert not task.done()
    assert task in zha_gateway._tracked_completable_tasks
    await task
    assert task not in zha_gateway._tracked_completable_tasks


async def test_async_add_zha_job_background(zha_gateway: Gateway) -> None:
    """Test scheduling a coro as a background task with async_add_zha_job."""

    async def job_that_suspends():
        await asyncio.sleep(0)

    task = zha_gateway.async_add_zha_job(ZHAJob(job_that_suspends), background=True)
    assert not task.done()
    assert task in zha_gateway._background_tasks
    await task
    assert task not in zha_gateway._background_tasks


async def test_async_run_zha_job_background(zha_gateway: Gateway) -> None:
    """Test scheduling a coro as a background task with async_run_zha_job."""

    async def job_that_suspends():
        await asyncio.sleep(0)

    task = zha_gateway.async_run_zha_job(ZHAJob(job_that_suspends), background=True)
    assert not task.done()
    assert task in zha_gateway._background_tasks
    await task
    assert task not in zha_gateway._background_tasks


async def test_async_add_zha_job_eager_background(zha_gateway: Gateway) -> None:
    """Test scheduling a coro as an eager background task with async_add_zha_job."""

    async def job_that_suspends():
        await asyncio.sleep(0)

    task = zha_gateway.async_add_zha_job(ZHAJob(job_that_suspends), background=True)
    assert not task.done()
    assert task in zha_gateway._background_tasks
    await task
    assert task not in zha_gateway._background_tasks


async def test_async_run_zha_job_eager_background(zha_gateway: Gateway) -> None:
    """Test scheduling a coro as an eager background task with async_run_zha_job."""

    async def job_that_suspends():
        await asyncio.sleep(0)

    task = zha_gateway.async_run_zha_job(ZHAJob(job_that_suspends), background=True)
    assert not task.done()
    assert task in zha_gateway._background_tasks
    await task
    assert task not in zha_gateway._background_tasks


@pytest.mark.parametrize("background", [True, False])
async def test_async_run_zha_job_background_no_suspend(
    zha_gateway: Gateway,
    background: bool,
) -> None:
    """Test scheduling a coro as an eager background task with async_run_zha_job."""

    async def job_that_does_not_suspends():
        pass

    task = zha_gateway.async_run_zha_job(
        ZHAJob(job_that_does_not_suspends),
        background=background,
    )
    assert task is not None
    assert task.done()
    assert task not in zha_gateway._background_tasks
    assert task not in zha_gateway._tracked_completable_tasks
    await task


async def test_async_add_zha_job_coro_named(zha_gateway: Gateway) -> None:
    """Test that we schedule coroutines and add jobs to the job pool with a name."""

    async def mycoro():
        pass

    job = ZHAJob(mycoro, "named coro")
    assert "named coro" in str(job)
    assert job.name == "named coro"
    task = AsyncUtilMixin.async_add_zha_job(zha_gateway, job)
    assert "named coro" in str(task)


async def test_async_add_zha_job_eager_start(zha_gateway: Gateway) -> None:
    """Test eager_start with async_add_zha_job."""

    async def mycoro():
        pass

    job = ZHAJob(mycoro, "named coro")
    assert "named coro" in str(job)
    assert job.name == "named coro"
    task = AsyncUtilMixin.async_add_zha_job(zha_gateway, job, eager_start=True)
    assert "named coro" in str(task)


async def test_async_add_zha_job_schedule_partial_callback() -> None:
    """Test that we schedule partial coros and add jobs to the job pool."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))
    job = MagicMock()
    partial = functools.partial(job)

    AsyncUtilMixin.async_add_zha_job(zha_gateway, ZHAJob(partial))
    assert len(zha_gateway.loop.call_soon.mock_calls) == 1
    assert len(zha_gateway.loop.create_task.mock_calls) == 0
    assert len(zha_gateway.add_job.mock_calls) == 0


async def test_async_add_zha_job_schedule_coroutinefunction() -> None:
    """Test that we schedule coroutines and add jobs to the job pool."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))

    async def job():
        pass

    AsyncUtilMixin.async_add_zha_job(zha_gateway, ZHAJob(job))
    assert len(zha_gateway.loop.call_soon.mock_calls) == 0
    assert len(zha_gateway.loop.create_task.mock_calls) == 1
    assert len(zha_gateway.add_job.mock_calls) == 0


async def test_async_add_zha_job_schedule_corofunction_eager_start() -> None:
    """Test that we schedule coroutines and add jobs to the job pool."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))

    async def job():
        pass

    with patch(
        "zha.async_.create_eager_task", wraps=create_eager_task
    ) as mock_create_eager_task:
        zha_job = ZHAJob(job)
        task = AsyncUtilMixin.async_add_zha_job(zha_gateway, zha_job, eager_start=True)
        assert task is not None
        assert len(zha_gateway.loop.call_soon.mock_calls) == 0
        assert len(zha_gateway.add_job.mock_calls) == 0
        assert mock_create_eager_task.mock_calls
        await task


async def test_async_add_zha_job_schedule_partial_coroutinefunction(
    zha_gateway: Gateway,
) -> None:
    """Test that we schedule partial coros and add jobs to the job pool."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))

    async def job():
        pass

    partial = functools.partial(job)

    AsyncUtilMixin.async_add_zha_job(zha_gateway, ZHAJob(partial))
    assert len(zha_gateway.loop.call_soon.mock_calls) == 0
    assert len(zha_gateway.loop.create_task.mock_calls) == 1
    assert len(zha_gateway.add_job.mock_calls) == 0


async def test_async_add_job_add_zha_threaded_job_to_pool() -> None:
    """Test that we schedule coroutines and add jobs to the job pool."""
    zha_gateway = MagicMock()

    def job():
        pass

    AsyncUtilMixin.async_add_zha_job(zha_gateway, ZHAJob(job))
    assert len(zha_gateway.loop.call_soon.mock_calls) == 1
    assert len(zha_gateway.loop.create_task.mock_calls) == 0
    assert len(zha_gateway.loop.run_in_executor.mock_calls) == 0


async def test_async_create_task_schedule_coroutine() -> None:
    """Test that we schedule coroutines and add jobs to the job pool."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))

    async def job():
        pass

    AsyncUtilMixin.async_create_task(zha_gateway, job())
    assert len(zha_gateway.loop.call_soon.mock_calls) == 0
    assert len(zha_gateway.loop.create_task.mock_calls) == 1
    assert len(zha_gateway.add_job.mock_calls) == 0


async def test_async_create_task_eager_start_schedule_coroutine() -> None:
    """Test that we schedule coroutines and add jobs to the job pool."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))

    async def job():
        pass

    AsyncUtilMixin.async_create_task(zha_gateway, job(), eager_start=True)
    # Should create the task directly since 3.12 supports eager_start
    assert len(zha_gateway.loop.create_task.mock_calls) == 0
    assert len(zha_gateway.add_job.mock_calls) == 0


async def test_async_create_task_schedule_coroutine_with_name() -> None:
    """Test that we schedule coroutines and add jobs to the job pool with a name."""
    zha_gateway = MagicMock(loop=MagicMock(wraps=asyncio.get_running_loop()))

    async def job():
        pass

    task = AsyncUtilMixin.async_create_task(zha_gateway, job(), "named task")
    assert len(zha_gateway.loop.call_soon.mock_calls) == 0
    assert len(zha_gateway.loop.create_task.mock_calls) == 1
    assert len(zha_gateway.add_job.mock_calls) == 0
    assert "named task" in str(task)


async def test_async_run_eager_zha_job_calls_callback() -> None:
    """Test that the callback annotation is respected."""
    zha_gateway = MagicMock()
    calls = []

    def job():
        asyncio.get_running_loop()  # ensure we are in the event loop
        calls.append(1)

    AsyncUtilMixin.async_run_zha_job(zha_gateway, ZHAJob(job))
    assert len(calls) == 1


async def test_async_run_eager_zha_job_calls_coro_function() -> None:
    """Test running coros from async_run_zha_job with eager_start."""
    zha_gateway = MagicMock()

    async def job():
        pass

    AsyncUtilMixin.async_run_zha_job(zha_gateway, ZHAJob(job))
    assert len(zha_gateway.async_add_zha_job.mock_calls) == 1


async def test_async_run_zha_job_calls_callback() -> None:
    """Test that the callback annotation is respected."""
    zha_gateway = MagicMock()
    calls = []

    def job():
        calls.append(1)

    AsyncUtilMixin.async_run_zha_job(zha_gateway, ZHAJob(job))
    assert len(calls) == 1
    assert len(zha_gateway.async_add_job.mock_calls) == 0


async def test_async_create_task_pending_tasks_coro(zha_gateway: Gateway) -> None:
    """Add a coro to pending tasks."""
    call_count = []

    async def test_coro():
        """Test Coro."""
        call_count.append("call")

    for _ in range(2):
        zha_gateway.async_create_task(test_coro())

    assert len(zha_gateway._tracked_completable_tasks) == 2
    await zha_gateway.async_block_till_done()
    assert len(call_count) == 2
    assert len(zha_gateway._tracked_completable_tasks) == 0


@pytest.mark.parametrize("eager_start", [True, False])
async def test_background_task(zha_gateway: Gateway, eager_start: bool) -> None:
    """Test background tasks being quit."""
    result = asyncio.get_running_loop().create_future()

    async def test_task():
        try:
            await asyncio.sleep(1)
        except asyncio.CancelledError:
            result.set_result("Foo")
            raise

    task = zha_gateway.async_create_background_task(
        test_task(), "happy task", eager_start=eager_start
    )
    assert "happy task" in str(task)
    await asyncio.sleep(0)
    await zha_gateway.shutdown()
    assert result.result() == "Foo"


def test_ZHAJob_passing_job_type():
    """Test passing the job type to ZHAJob when we already know it."""

    def callback_func():
        pass

    def not_callback_func():
        pass

    assert (
        ZHAJob(callback_func, job_type=ZHAJobType.Callback).job_type
        == ZHAJobType.Callback
    )

    # We should trust the job_type passed in
    assert (
        ZHAJob(not_callback_func, job_type=ZHAJobType.Callback).job_type
        == ZHAJobType.Callback
    )


async def test_async_add_executor_job_background(zha_gateway: Gateway) -> None:
    """Test running an executor job in the background."""
    calls = []

    def job():
        time.sleep(0.01)
        calls.append(1)

    async def _async_add_executor_job():
        await zha_gateway.async_add_executor_job(job)

    task = zha_gateway.async_create_background_task(
        _async_add_executor_job(), "background", eager_start=True
    )
    await zha_gateway.async_block_till_done()
    assert len(calls) == 0
    await zha_gateway.async_block_till_done(wait_background_tasks=True)
    assert len(calls) == 1
    await task


async def test_async_add_executor_job(zha_gateway: Gateway) -> None:
    """Test running an executor job."""
    calls = []

    def job():
        time.sleep(0.01)
        calls.append(1)

    async def _async_add_executor_job():
        await zha_gateway.async_add_executor_job(job)

    task = zha_gateway.async_create_task(
        _async_add_executor_job(), "background", eager_start=True
    )
    await zha_gateway.async_block_till_done()
    assert len(calls) == 1
    await task


async def test_gather_with_limited_concurrency() -> None:
    """Test gather_with_limited_concurrency limits the number of running tasks."""

    runs = 0
    now_time = time.time()

    async def _increment_runs_if_in_time():
        if time.time() - now_time > 0.1:
            return -1

        nonlocal runs
        runs += 1
        await asyncio.sleep(0.1)
        return runs

    results = await zha_async.gather_with_limited_concurrency(
        2, *(_increment_runs_if_in_time() for i in range(4))
    )

    assert results == [2, 2, -1, -1]


async def test_create_eager_task_312(zha_gateway: Gateway) -> None:  # pylint: disable=unused-argument
    """Test create_eager_task schedules a task eagerly in the event loop.

    For Python 3.12+, the task is scheduled eagerly in the event loop.
    """
    events = []

    async def _normal_task():
        events.append("normal")

    async def _eager_task():
        events.append("eager")

    task1 = zha_async.create_eager_task(_eager_task())
    task2 = asyncio.create_task(_normal_task())

    assert events == ["eager"]

    await asyncio.sleep(0)
    assert events == ["eager", "normal"]
    await task1
    await task2



"""Test zha switch."""

import asyncio
import logging
from unittest.mock import call, patch

import pytest
from zhaquirks.const import (
    DEVICE_TYPE,
    ENDPOINTS,
    INPUT_CLUSTERS,
    OUTPUT_CLUSTERS,
    PROFILE_ID,
)
from zigpy.exceptions import ZigbeeException
from zigpy.profiles import zha
from zigpy.quirks import DEVICE_REGISTRY, CustomCluster, CustomDevice
from zigpy.quirks.v2 import CustomDeviceV2, QuirkBuilder
import zigpy.types as t
from zigpy.typing import UNDEFINED
from zigpy.zcl.clusters import closures, general
from zigpy.zcl.clusters.general import BinaryOutput
from zigpy.zcl.clusters.manufacturer_specific import ManufacturerSpecificCluster
import zigpy.zcl.foundation as zcl_f

from tests.common import (
    SIG_EP_INPUT,
    SIG_EP_OUTPUT,
    SIG_EP_PROFILE,
    SIG_EP_TYPE,
    create_mock_zigpy_device,
    get_entity,
    get_group_entity,
    group_entity_availability_test,
    join_zigpy_device,
    send_attributes_report,
    update_attribute_cache,
    zigpy_device_from_json,
)
from zha.application import Platform
from zha.application.gateway import Gateway
from zha.application.platforms import GroupEntity, PlatformEntity
from zha.exceptions import ZHAException
from zha.zigbee.device import Device
from zha.zigbee.group import Group, GroupMemberReference

ON = 1
OFF = 0
IEEE_GROUPABLE_DEVICE = "01:2d:6f:00:0a:90:69:e8"
IEEE_GROUPABLE_DEVICE2 = "02:2d:6f:00:0a:90:69:e8"
_LOGGER = logging.getLogger(__name__)


ZIGPY_DEVICE = {
    1: {
        SIG_EP_INPUT: [general.Basic.cluster_id, general.OnOff.cluster_id],
        SIG_EP_OUTPUT: [],
        SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
        SIG_EP_PROFILE: zha.PROFILE_ID,
    }
}


ZIGPY_COVER_DEVICE = {
    1: {
        SIG_EP_PROFILE: zha.PROFILE_ID,
        SIG_EP_TYPE: zha.DeviceType.WINDOW_COVERING_DEVICE,
        SIG_EP_INPUT: [
            general.Basic.cluster_id,
            closures.WindowCovering.cluster_id,
        ],
        SIG_EP_OUTPUT: [],
    }
}


async def device_switch_1_mock(zha_gateway: Gateway) -> Device:
    """Test zha switch platform."""

    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [general.OnOff.cluster_id, general.Groups.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        ieee=IEEE_GROUPABLE_DEVICE,
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    return zha_device


async def device_switch_2_mock(zha_gateway: Gateway) -> Device:
    """Test zha switch platform."""

    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [general.OnOff.cluster_id, general.Groups.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        ieee=IEEE_GROUPABLE_DEVICE2,
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    return zha_device


async def test_switch(zha_gateway: Gateway) -> None:
    """Test zha switch platform."""
    zigpy_device = create_mock_zigpy_device(zha_gateway, ZIGPY_DEVICE)
    zigpy_device.node_desc.mac_capability_flags |= (
        0b_0000_0100  # this one is mains powered
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    cluster = zigpy_device.endpoints.get(1).on_off
    entity: PlatformEntity = get_entity(zha_device, Platform.SWITCH)

    assert bool(bool(entity.state["state"])) is False

    # turn on at switch
    await send_attributes_report(zha_gateway, cluster, {1: 0, 0: 1, 2: 2})
    assert bool(entity.state["state"]) is True

    # turn off at switch
    await send_attributes_report(zha_gateway, cluster, {1: 1, 0: 0, 2: 2})
    assert bool(entity.state["state"]) is False

    # turn on from client
    with patch(
        "zigpy.zcl.Cluster.request",
        return_value=[0x00, zcl_f.Status.SUCCESS],
    ):
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert bool(entity.state["state"]) is True
        assert len(cluster.request.mock_calls) == 1
        assert cluster.request.call_args == call(
            False,
            ON,
            cluster.commands_by_name["on"].schema,
            expect_reply=True,
            manufacturer=None,
        )

    # Fail turn off from client
    with (
        patch(
            "zigpy.zcl.Cluster.request",
            return_value=[0x01, zcl_f.Status.FAILURE],
        ),
        pytest.raises(ZHAException, match="Failed to turn off"),
    ):
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert bool(entity.state["state"]) is True
        assert len(cluster.request.mock_calls) == 1
        assert cluster.request.call_args == call(
            False,
            OFF,
            cluster.commands_by_name["off"].schema,
            expect_reply=True,
            manufacturer=None,
        )

    # turn off from client
    with patch(
        "zigpy.zcl.Cluster.request",
        return_value=[0x01, zcl_f.Status.SUCCESS],
    ):
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert bool(entity.state["state"]) is False
        assert len(cluster.request.mock_calls) == 1
        assert cluster.request.call_args == call(
            False,
            OFF,
            cluster.commands_by_name["off"].schema,
            expect_reply=True,
            manufacturer=None,
        )

    # Fail turn on from client
    with (
        patch(
            "zigpy.zcl.Cluster.request",
            return_value=[0x01, zcl_f.Status.FAILURE],
        ),
        pytest.raises(ZHAException, match="Failed to turn on"),
    ):
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert bool(entity.state["state"]) is False
        assert len(cluster.request.mock_calls) == 1
        assert cluster.request.call_args == call(
            False,
            ON,
            cluster.commands_by_name["on"].schema,
            expect_reply=True,
            manufacturer=None,
        )

    # test updating entity state from client
    cluster.read_attributes.reset_mock()
    assert bool(entity.state["state"]) is False
    cluster.PLUGGED_ATTR_READS = {"on_off": True}
    await entity.async_update()
    await zha_gateway.async_block_till_done()
    assert cluster.read_attributes.await_count == 1
    assert cluster.read_attributes.await_args == call(
        ["on_off"], allow_cache=False, only_cache=False, manufacturer=UNDEFINED
    )
    assert bool(entity.state["state"]) is True


async def test_zha_group_switch_entity(zha_gateway: Gateway) -> None:
    """Test the switch entity for a ZHA group."""
    device_switch_1 = await device_switch_1_mock(zha_gateway)
    device_switch_2 = await device_switch_2_mock(zha_gateway)
    member_ieee_addresses = [device_switch_1.ieee, device_switch_2.ieee]
    members = [
        GroupMemberReference(ieee=device_switch_1.ieee, endpoint_id=1),
        GroupMemberReference(ieee=device_switch_2.ieee, endpoint_id=1),
    ]

    # test creating a group with 2 members
    zha_group: Group = await zha_gateway.async_create_zigpy_group("Test Group", members)
    await zha_gateway.async_block_till_done()

    assert zha_group is not None
    assert len(zha_group.members) == 2
    for member in zha_group.members:
        assert member.device.ieee in member_ieee_addresses
        assert member.group == zha_group
        assert member.endpoint is not None

    entity: GroupEntity = get_group_entity(zha_group, platform=Platform.SWITCH)
    assert entity.group_id == zha_group.group_id
    assert entity.info_object.fallback_name == zha_group.name

    group_cluster_on_off = zha_group.zigpy_group.endpoint[general.OnOff.cluster_id]
    dev1_cluster_on_off = device_switch_1.device.endpoints[1].on_off
    dev2_cluster_on_off = device_switch_2.device.endpoints[1].on_off

    # test that the lights were created and are off
    assert bool(entity.state["state"]) is False

    # turn on from HA
    with patch(
        "zigpy.zcl.Cluster.request",
        return_value=[0x00, zcl_f.Status.SUCCESS],
    ):
        # turn on via UI
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert len(group_cluster_on_off.request.mock_calls) == 1
        assert group_cluster_on_off.request.call_args == call(
            False,
            ON,
            group_cluster_on_off.commands_by_name["on"].schema,
            expect_reply=True,
            manufacturer=None,
        )
    assert bool(entity.state["state"]) is True

    # turn off from HA
    with patch(
        "zigpy.zcl.Cluster.request",
        return_value=[0x01, zcl_f.Status.SUCCESS],
    ):
        # turn off via UI
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert len(group_cluster_on_off.request.mock_calls) == 1
        assert group_cluster_on_off.request.call_args == call(
            False,
            OFF,
            group_cluster_on_off.commands_by_name["off"].schema,
            expect_reply=True,
            manufacturer=None,
        )
    assert bool(entity.state["state"]) is False

    # test some of the group logic to make sure we key off states correctly
    await send_attributes_report(zha_gateway, dev1_cluster_on_off, {0: 1})
    await send_attributes_report(zha_gateway, dev2_cluster_on_off, {0: 1})
    await zha_gateway.async_block_till_done()

    # group member updates are debounced
    assert bool(entity.state["state"]) is False
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()

    # test that group light is on
    assert bool(entity.state["state"]) is True

    await send_attributes_report(zha_gateway, dev1_cluster_on_off, {0: 0})
    await zha_gateway.async_block_till_done()

    # test that group light is still on
    assert bool(entity.state["state"]) is True

    await send_attributes_report(zha_gateway, dev2_cluster_on_off, {0: 0})
    await zha_gateway.async_block_till_done()

    # group member updates are debounced
    assert bool(entity.state["state"]) is True
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()

    # test that group light is now off
    assert bool(entity.state["state"]) is False

    await send_attributes_report(zha_gateway, dev1_cluster_on_off, {0: 1})
    await zha_gateway.async_block_till_done()

    # group member updates are debounced
    assert bool(entity.state["state"]) is False
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()

    # test that group light is now back on
    assert bool(entity.state["state"]) is True

    await group_entity_availability_test(
        zha_gateway, device_switch_1, device_switch_2, entity
    )


class WindowDetectionFunctionQuirk(CustomDevice):
    """Quirk with window detection function attribute."""

    class TuyaManufCluster(CustomCluster, ManufacturerSpecificCluster):
        """Tuya manufacturer specific cluster."""

        cluster_id = 0xEF00
        ep_attribute = "tuya_manufacturer"

        attributes = {
            0xEF01: ("window_detection_function", t.Bool),
            0xEF02: ("window_detection_function_inverter", t.Bool),
        }

        def __init__(self, *args, **kwargs):
            """Initialize with task."""
            super().__init__(*args, **kwargs)
            self._attr_cache.update(
                {0xEF01: False}
            )  # entity won't be created without this

    replacement = {
        ENDPOINTS: {
            1: {
                PROFILE_ID: zha.PROFILE_ID,
                DEVICE_TYPE: zha.DeviceType.ON_OFF_SWITCH,
                INPUT_CLUSTERS: [general.Basic.cluster_id, TuyaManufCluster],
                OUTPUT_CLUSTERS: [],
            },
        }
    }


async def test_switch_configurable(
    zha_gateway: Gateway,
) -> None:
    """Test ZHA configurable switch platform."""

    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [general.Basic.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        manufacturer="_TZE200_b6wax7g0",
        quirk=WindowDetectionFunctionQuirk,
    )
    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)

    cluster = zigpy_dev.endpoints[1].tuya_manufacturer
    entity = get_entity(zha_device, platform=Platform.SWITCH)

    # test that the state has changed from unavailable to off
    assert bool(entity.state["state"]) is False

    # turn on at switch
    await send_attributes_report(
        zha_gateway, cluster, {"window_detection_function": True}
    )
    assert bool(entity.state["state"]) is True

    # turn off at switch
    await send_attributes_report(
        zha_gateway, cluster, {"window_detection_function": False}
    )
    assert bool(entity.state["state"]) is False

    # turn on from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.Status.SUCCESS, zcl_f.Status.SUCCESS],
    ):
        # turn on via UI
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": True}, manufacturer=UNDEFINED)
        ]

    cluster.write_attributes.reset_mock()

    # turn off from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.Status.SUCCESS, zcl_f.Status.SUCCESS],
    ):
        # turn off via UI
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": False}, manufacturer=UNDEFINED)
        ]

    cluster.read_attributes.reset_mock()
    await entity.async_update()
    await zha_gateway.async_block_till_done()
    # the mocking doesn't update the attr cache so this flips back to initial value
    assert cluster.read_attributes.call_count == 1
    assert [
        call(
            [
                "window_detection_function",
                "window_detection_function_inverter",
            ],
            allow_cache=False,
            only_cache=False,
            manufacturer=UNDEFINED,
        )
    ] == cluster.read_attributes.call_args_list

    cluster.write_attributes.reset_mock()
    cluster.write_attributes.side_effect = ZigbeeException

    with pytest.raises(ZHAException):
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()

    assert cluster.write_attributes.mock_calls == [
        call({"window_detection_function": False}, manufacturer=UNDEFINED),
        call({"window_detection_function": False}, manufacturer=UNDEFINED),
        call({"window_detection_function": False}, manufacturer=UNDEFINED),
    ]

    cluster.write_attributes.side_effect = None

    # test inverter
    cluster.write_attributes.reset_mock()
    cluster._attr_cache.update({0xEF02: True})

    await entity.async_turn_off()
    await zha_gateway.async_block_till_done()
    assert cluster.write_attributes.mock_calls == [
        call({"window_detection_function": True}, manufacturer=UNDEFINED)
    ]

    cluster.write_attributes.reset_mock()
    await entity.async_turn_on()
    await zha_gateway.async_block_till_done()
    assert cluster.write_attributes.mock_calls == [
        call({"window_detection_function": False}, manufacturer=UNDEFINED)
    ]


async def test_switch_configurable_custom_on_off_values(zha_gateway: Gateway) -> None:
    """Test ZHA configurable switch platform."""

    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [general.Basic.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        manufacturer="manufacturer",
        model="model",
    )

    (
        QuirkBuilder(zigpy_dev.manufacturer, zigpy_dev.model)
        .adds(WindowDetectionFunctionQuirk.TuyaManufCluster)
        .switch(
            "window_detection_function",
            WindowDetectionFunctionQuirk.TuyaManufCluster.cluster_id,
            on_value=3,
            off_value=5,
            translation_key="window_detection_function",
            fallback_name="Window detection function",
        )
        .add_to_registry()
    )

    zigpy_device_ = DEVICE_REGISTRY.get_device(zigpy_dev)

    assert isinstance(zigpy_device_, CustomDeviceV2)
    cluster = zigpy_device_.endpoints[1].tuya_manufacturer
    cluster.PLUGGED_ATTR_READS = {"window_detection_function": 5}
    update_attribute_cache(cluster)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device_)

    entity = get_entity(zha_device, platform=Platform.SWITCH)

    assert bool(entity.state["state"]) is False

    # turn on at switch
    await send_attributes_report(zha_gateway, cluster, {"window_detection_function": 3})
    assert bool(entity.state["state"]) is True

    # turn off at switch
    await send_attributes_report(zha_gateway, cluster, {"window_detection_function": 5})
    assert bool(entity.state["state"]) is False

    # turn on from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.WriteAttributesResponse.deserialize(b"\x00")[0]],
    ):
        # turn on via UI
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": 3}, manufacturer=UNDEFINED)
        ]
        cluster.write_attributes.reset_mock()

    # turn off from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.WriteAttributesResponse.deserialize(b"\x00")[0]],
    ):
        # turn off via UI
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": 5}, manufacturer=UNDEFINED)
        ]


async def test_switch_configurable_custom_on_off_values_force_inverted(
    zha_gateway: Gateway,
) -> None:
    """Test ZHA configurable switch platform."""

    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [general.Basic.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        manufacturer="manufacturer2",
        model="model2",
    )

    (
        QuirkBuilder(zigpy_dev.manufacturer, zigpy_dev.model)
        .adds(WindowDetectionFunctionQuirk.TuyaManufCluster)
        .switch(
            "window_detection_function",
            WindowDetectionFunctionQuirk.TuyaManufCluster.cluster_id,
            on_value=3,
            off_value=5,
            force_inverted=True,
            translation_key="window_detection_function",
            fallback_name="Window detection function",
        )
        .add_to_registry()
    )

    zigpy_device_ = DEVICE_REGISTRY.get_device(zigpy_dev)

    assert isinstance(zigpy_device_, CustomDeviceV2)
    cluster = zigpy_device_.endpoints[1].tuya_manufacturer
    cluster.PLUGGED_ATTR_READS = {"window_detection_function": 5}
    update_attribute_cache(cluster)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device_)

    entity = get_entity(zha_device, platform=Platform.SWITCH)

    assert bool(entity.state["state"]) is True

    # turn on at switch
    await send_attributes_report(zha_gateway, cluster, {"window_detection_function": 3})
    assert bool(entity.state["state"]) is False

    # turn off at switch
    await send_attributes_report(zha_gateway, cluster, {"window_detection_function": 5})
    assert bool(entity.state["state"]) is True

    # turn on from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.WriteAttributesResponse.deserialize(b"\x00")[0]],
    ):
        # turn on via UI
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": 5}, manufacturer=UNDEFINED)
        ]
        cluster.write_attributes.reset_mock()

    # turn off from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.WriteAttributesResponse.deserialize(b"\x00")[0]],
    ):
        # turn off via UI
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": 3}, manufacturer=UNDEFINED)
        ]


async def test_switch_configurable_custom_on_off_values_inverter_attribute(
    zha_gateway: Gateway,
) -> None:
    """Test ZHA configurable switch platform."""

    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [general.Basic.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zha.PROFILE_ID,
            }
        },
        manufacturer="manufacturer3",
        model="model3",
    )

    (
        QuirkBuilder(zigpy_dev.manufacturer, zigpy_dev.model)
        .adds(WindowDetectionFunctionQuirk.TuyaManufCluster)
        .switch(
            "window_detection_function",
            WindowDetectionFunctionQuirk.TuyaManufCluster.cluster_id,
            on_value=3,
            off_value=5,
            invert_attribute_name="window_detection_function_inverter",
            translation_key="window_detection_function",
            fallback_name="Window detection function",
        )
        .add_to_registry()
    )

    zigpy_device_ = DEVICE_REGISTRY.get_device(zigpy_dev)

    assert isinstance(zigpy_device_, CustomDeviceV2)
    cluster = zigpy_device_.endpoints[1].tuya_manufacturer
    cluster.PLUGGED_ATTR_READS = {
        "window_detection_function": 5,
        "window_detection_function_inverter": t.Bool(True),
    }
    update_attribute_cache(cluster)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device_)

    entity = get_entity(zha_device, platform=Platform.SWITCH)

    assert bool(entity.state["state"]) is True

    # turn on at switch
    await send_attributes_report(zha_gateway, cluster, {"window_detection_function": 3})
    assert bool(entity.state["state"]) is False

    # turn off at switch
    await send_attributes_report(zha_gateway, cluster, {"window_detection_function": 5})
    assert bool(entity.state["state"]) is True

    # turn on from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.WriteAttributesResponse.deserialize(b"\x00")[0]],
    ):
        # turn on via UI
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": 5}, manufacturer=UNDEFINED)
        ]
        cluster.write_attributes.reset_mock()

    # turn off from HA
    with patch(
        "zigpy.zcl.Cluster.write_attributes",
        return_value=[zcl_f.WriteAttributesResponse.deserialize(b"\x00")[0]],
    ):
        # turn off via UI
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.mock_calls == [
            call({"window_detection_function": 3}, manufacturer=UNDEFINED)
        ]


WCAttrs = closures.WindowCovering.AttributeDefs
WCT = closures.WindowCovering.WindowCoveringType
WCCS = closures.WindowCovering.ConfigStatus
WCM = closures.WindowCovering.WindowCoveringMode


async def test_cover_inversion_switch(zha_gateway: Gateway) -> None:
    """Test ZHA cover platform."""

    # load up cover domain
    zigpy_cover_device = create_mock_zigpy_device(zha_gateway, ZIGPY_COVER_DEVICE)
    cluster = zigpy_cover_device.endpoints[1].window_covering
    cluster.PLUGGED_ATTR_READS = {
        WCAttrs.current_position_lift_percentage.name: 65,
        WCAttrs.current_position_tilt_percentage.name: 42,
        WCAttrs.window_covering_type.name: WCT.Tilt_blind_tilt_and_lift,
        WCAttrs.config_status.name: WCCS(~WCCS.Open_up_commands_reversed),
        WCAttrs.window_covering_mode.name: WCM(WCM.LEDs_display_feedback),
    }
    update_attribute_cache(cluster)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_cover_device)
    assert (
        not zha_device.endpoints[1]
        .all_cluster_handlers[f"1:0x{cluster.cluster_id:04x}"]
        .inverted
    )
    assert cluster.read_attributes.call_count == 3
    assert (
        WCAttrs.current_position_lift_percentage.name
        in cluster.read_attributes.call_args[0][0]
    )
    assert (
        WCAttrs.current_position_tilt_percentage.name
        in cluster.read_attributes.call_args[0][0]
    )

    entity = get_entity(zha_device, platform=Platform.SWITCH)

    # test update
    prev_call_count = cluster.read_attributes.call_count
    await entity.async_update()
    await zha_gateway.async_block_till_done()
    assert cluster.read_attributes.call_count == prev_call_count + 1
    assert bool(entity.state["state"]) is False

    # test to see the state remains after tilting to 0%
    await send_attributes_report(
        zha_gateway, cluster, {WCAttrs.current_position_tilt_percentage.id: 0}
    )
    assert bool(entity.state["state"]) is False

    with patch(
        "zigpy.zcl.Cluster.write_attributes", return_value=[0x1, zcl_f.Status.SUCCESS]
    ):
        cluster.PLUGGED_ATTR_READS = {
            WCAttrs.config_status.name: WCCS.Operational
            | WCCS.Open_up_commands_reversed,
        }
        # turn on from UI
        await entity.async_turn_on()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.call_count == 1
        assert cluster.write_attributes.call_args_list[0] == call(
            {
                WCAttrs.window_covering_mode.name: WCM.Motor_direction_reversed
                | WCM.LEDs_display_feedback
            },
            manufacturer=UNDEFINED,
        )

        assert bool(entity.state["state"]) is True

        cluster.write_attributes.reset_mock()

        # turn off from UI
        cluster.PLUGGED_ATTR_READS = {
            WCAttrs.config_status.name: WCCS.Operational,
        }
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.call_count == 1
        assert cluster.write_attributes.call_args_list[0] == call(
            {WCAttrs.window_covering_mode.name: WCM.LEDs_display_feedback},
            manufacturer=UNDEFINED,
        )

        assert bool(entity.state["state"]) is False

        cluster.write_attributes.reset_mock()

        # test that sending the command again does not result in a write
        await entity.async_turn_off()
        await zha_gateway.async_block_till_done()
        assert cluster.write_attributes.call_count == 0

        assert bool(entity.state["state"]) is False


async def test_cover_inversion_switch_not_created(zha_gateway: Gateway) -> None:
    """Test ZHA cover platform."""

    # load up cover domain
    zigpy_cover_device = create_mock_zigpy_device(zha_gateway, ZIGPY_COVER_DEVICE)
    cluster = zigpy_cover_device.endpoints[1].window_covering
    cluster.PLUGGED_ATTR_READS = {
        WCAttrs.current_position_lift_percentage.name: 65,
        WCAttrs.current_position_tilt_percentage.name: 42,
        WCAttrs.config_status.name: WCCS(~WCCS.Open_up_commands_reversed),
    }
    update_attribute_cache(cluster)
    zha_device = await join_zigpy_device(zha_gateway, zigpy_cover_device)

    assert cluster.read_attributes.call_count == 3
    assert (
        WCAttrs.current_position_lift_percentage.name
        in cluster.read_attributes.call_args[0][0]
    )
    assert (
        WCAttrs.current_position_tilt_percentage.name
        in cluster.read_attributes.call_args[0][0]
    )

    # entity should not be created when mode or config status aren't present
    with pytest.raises(KeyError):
        get_entity(zha_device, platform=Platform.SWITCH)


async def test_binary_output_cluster(zha_gateway: Gateway) -> None:
    """Test ZHA switch platform with binary output cluster."""

    zigpy_device = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/espressif-zigbeebinaryoutputdevice.json",
    )
    cluster = zigpy_device.endpoints[1].binary_output
    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    switch_entity = get_entity(zha_device, platform=Platform.SWITCH)

    # Clear out the attribute first, to test handling of the missing state
    cluster.update_attribute(BinaryOutput.AttributeDefs.present_value.id, None)

    assert switch_entity.info_object.fallback_name == "Entity Description"
    assert switch_entity.state["state"] is False

    # Turn it on
    cluster.write_attributes.reset_mock()
    await switch_entity.async_turn_on()
    assert cluster.write_attributes.mock_calls == [
        call({"present_value": True}, manufacturer=UNDEFINED)
    ]
    assert switch_entity.state["state"] is True

    # Turn it off
    cluster.write_attributes.reset_mock()
    await switch_entity.async_turn_off()
    assert cluster.write_attributes.mock_calls == [
        call({"present_value": False}, manufacturer=UNDEFINED)
    ]
    assert switch_entity.state["state"] is False

    # Report an attribute change
    await send_attributes_report(
        zha_gateway,
        cluster,
        {BinaryOutput.AttributeDefs.present_value.id: t.Bool(False)},
    )
    assert switch_entity.state["state"] is False

    # Force an update
    cluster.read_attributes.reset_mock()
    cluster.PLUGGED_ATTR_READS = {BinaryOutput.AttributeDefs.present_value.name: True}

    await switch_entity.async_update()
    assert switch_entity.state["state"] is True

    assert cluster.read_attributes.mock_calls == [
        call(
            [BinaryOutput.AttributeDefs.present_value.name],
            allow_cache=False,
            only_cache=False,
            manufacturer=UNDEFINED,
        )
    ]



"""Constants for the sensor platform."""

from datetime import UTC, datetime
import enum

from zigpy.zcl.clusters.general_const import AnalogInputType

from zha.units import (
    CONCENTRATION_PARTS_PER_MILLION,
    COUNT,
    KILOJOULES_PER_KG,
    PERCENTAGE,
    REVOLUTIONS_PER_MINUTE,
    UnitOfElectricCurrent,
    UnitOfEnergy,
    UnitOfFrequency,
    UnitOfPower,
    UnitOfPressure,
    UnitOfTemperature,
    UnitOfTime,
)


class SensorStateClass(enum.StrEnum):
    """State class for sensors."""

    MEASUREMENT = "measurement"
    """The state represents a measurement in present time."""

    TOTAL = "total"
    """The state represents a total amount.

    For example: net energy consumption"""

    TOTAL_INCREASING = "total_increasing"
    """The state represents a monotonically increasing total.

    For example: an amount of consumed gas"""


class SensorDeviceClass(enum.StrEnum):
    """Device class for sensors."""

    # Non-numerical device classes
    DATE = "date"
    """Date.

    Unit of measurement: `None`

    ISO8601 format: https://en.wikipedia.org/wiki/ISO_8601
    """

    ENUM = "enum"
    """Enumeration.

    Provides a fixed list of options the state of the sensor can be in.

    Unit of measurement: `None`
    """

    TIMESTAMP = "timestamp"
    """Timestamp.

    Unit of measurement: `None`

    ISO8601 format: https://en.wikipedia.org/wiki/ISO_8601
    """

    # Numerical device classes, these should be aligned with NumberDeviceClass
    APPARENT_POWER = "apparent_power"
    """Apparent power.

    Unit of measurement: `VA`
    """

    AQI = "aqi"
    """Air Quality Index.

    Unit of measurement: `None`
    """

    ATMOSPHERIC_PRESSURE = "atmospheric_pressure"
    """Atmospheric pressure.

    Unit of measurement: `UnitOfPressure` units
    """

    BATTERY = "battery"
    """Percentage of battery that is left.

    Unit of measurement: `%`
    """

    CO = "carbon_monoxide"
    """Carbon Monoxide gas concentration.

    Unit of measurement: `ppm` (parts per million)
    """

    CO2 = "carbon_dioxide"
    """Carbon Dioxide gas concentration.

    Unit of measurement: `ppm` (parts per million)
    """

    CONDUCTIVITY = "conductivity"
    """Conductivity.

    Unit of measurement: `S/cm`, `mS/cm`, `µS/cm`
    """

    CURRENT = "current"
    """Current.

    Unit of measurement: `A`, `mA`
    """

    DATA_RATE = "data_rate"
    """Data rate.

    Unit of measurement: UnitOfDataRate
    """

    DATA_SIZE = "data_size"
    """Data size.

    Unit of measurement: UnitOfInformation
    """

    DISTANCE = "distance"
    """Generic distance.

    Unit of measurement: `LENGTH_*` units
    - SI /metric: `mm`, `cm`, `m`, `km`
    - USCS / imperial: `in`, `ft`, `yd`, `mi`
    """

    DURATION = "duration"
    """Fixed duration.

    Unit of measurement: `d`, `h`, `min`, `s`, `ms`
    """

    ENERGY = "energy"
    """Energy.

    Use this device class for sensors measuring energy consumption, for example
    electric energy consumption.
    Unit of measurement: `Wh`, `kWh`, `MWh`, `MJ`, `GJ`
    """

    ENERGY_STORAGE = "energy_storage"
    """Stored energy.

    Use this device class for sensors measuring stored energy, for example the amount
    of electric energy currently stored in a battery or the capacity of a battery.

    Unit of measurement: `Wh`, `kWh`, `MWh`, `MJ`, `GJ`
    """

    FREQUENCY = "frequency"
    """Frequency.

    Unit of measurement: `Hz`, `kHz`, `MHz`, `GHz`
    """

    GAS = "gas"
    """Gas.

    Unit of measurement:
    - SI / metric: `m³`
    - USCS / imperial: `ft³`, `CCF`
    """

    HUMIDITY = "humidity"
    """Relative humidity.

    Unit of measurement: `%`
    """

    ILLUMINANCE = "illuminance"
    """Illuminance.

    Unit of measurement: `lx`
    """

    IRRADIANCE = "irradiance"
    """Irradiance.

    Unit of measurement:
    - SI / metric: `W/m²`
    - USCS / imperial: `BTU/(h⋅ft²)`
    """

    MOISTURE = "moisture"
    """Moisture.

    Unit of measurement: `%`
    """

    MONETARY = "monetary"
    """Amount of money.

    Unit of measurement: ISO4217 currency code

    See https://en.wikipedia.org/wiki/ISO_4217#Active_codes for active codes
    """

    NITROGEN_DIOXIDE = "nitrogen_dioxide"
    """Amount of NO2.

    Unit of measurement: `µg/m³`
    """

    NITROGEN_MONOXIDE = "nitrogen_monoxide"
    """Amount of NO.

    Unit of measurement: `µg/m³`
    """

    NITROUS_OXIDE = "nitrous_oxide"
    """Amount of N2O.

    Unit of measurement: `µg/m³`
    """

    OZONE = "ozone"
    """Amount of O3.

    Unit of measurement: `µg/m³`
    """

    PH = "ph"
    """Potential hydrogen (acidity/alkalinity).

    Unit of measurement: Unitless
    """

    PM1 = "pm1"
    """Particulate matter <= 1 μm.

    Unit of measurement: `µg/m³`
    """

    PM10 = "pm10"
    """Particulate matter <= 10 μm.

    Unit of measurement: `µg/m³`
    """

    PM25 = "pm25"
    """Particulate matter <= 2.5 μm.

    Unit of measurement: `µg/m³`
    """

    POWER_FACTOR = "power_factor"
    """Power factor.

    Unit of measurement: `%`, `None`
    """

    POWER = "power"
    """Power.

    Unit of measurement: `W`, `kW`
    """

    PRECIPITATION = "precipitation"
    """Accumulated precipitation.

    Unit of measurement: UnitOfPrecipitationDepth
    - SI / metric: `cm`, `mm`
    - USCS / imperial: `in`
    """

    PRECIPITATION_INTENSITY = "precipitation_intensity"
    """Precipitation intensity.

    Unit of measurement: UnitOfVolumetricFlux
    - SI /metric: `mm/d`, `mm/h`
    - USCS / imperial: `in/d`, `in/h`
    """

    PRESSURE = "pressure"
    """Pressure.

    Unit of measurement:
    - `mbar`, `cbar`, `bar`
    - `Pa`, `hPa`, `kPa`
    - `inHg`
    - `psi`
    """

    REACTIVE_POWER = "reactive_power"
    """Reactive power.

    Unit of measurement: `var`
    """

    SIGNAL_STRENGTH = "signal_strength"
    """Signal strength.

    Unit of measurement: `dB`, `dBm`
    """

    SOUND_PRESSURE = "sound_pressure"
    """Sound pressure.

    Unit of measurement: `dB`, `dBA`
    """

    SPEED = "speed"
    """Generic speed.

    Unit of measurement: `SPEED_*` units or `UnitOfVolumetricFlux`
    - SI /metric: `mm/d`, `mm/h`, `m/s`, `km/h`
    - USCS / imperial: `in/d`, `in/h`, `ft/s`, `mph`
    - Nautical: `kn`
    - Beaufort: `Beaufort`
    """

    SULPHUR_DIOXIDE = "sulphur_dioxide"
    """Amount of SO2.

    Unit of measurement: `µg/m³`
    """

    TEMPERATURE = "temperature"
    """Temperature.

    Unit of measurement: `°C`, `°F`, `K`
    """

    VOLATILE_ORGANIC_COMPOUNDS = "volatile_organic_compounds"
    """Amount of VOC.

    Unit of measurement: `µg/m³`
    """

    VOLATILE_ORGANIC_COMPOUNDS_PARTS = "volatile_organic_compounds_parts"
    """Ratio of VOC.

    Unit of measurement: `ppm`, `ppb`
    """

    VOLTAGE = "voltage"
    """Voltage.

    Unit of measurement: `V`, `mV`
    """

    VOLUME = "volume"
    """Generic volume.

    Unit of measurement: `VOLUME_*` units
    - SI / metric: `mL`, `L`, `m³`
    - USCS / imperial: `ft³`, `CCF`, `fl. oz.`, `gal` (warning: volumes expressed in
    USCS/imperial units are currently assumed to be US volumes)
    """

    VOLUME_STORAGE = "volume_storage"
    """Generic stored volume.

    Use this device class for sensors measuring stored volume, for example the amount
    of fuel in a fuel tank.

    Unit of measurement: `VOLUME_*` units
    - SI / metric: `mL`, `L`, `m³`
    - USCS / imperial: `ft³`, `CCF`, `fl. oz.`, `gal` (warning: volumes expressed in
    USCS/imperial units are currently assumed to be US volumes)
    """

    VOLUME_FLOW_RATE = "volume_flow_rate"
    """Generic flow rate

    Unit of measurement: UnitOfVolumeFlowRate
    - SI / metric: `m³/h`, `L/min`
    - USCS / imperial: `ft³/min`, `gal/min`
    """

    WATER = "water"
    """Water.

    Unit of measurement:
    - SI / metric: `m³`, `L`
    - USCS / imperial: `ft³`, `CCF`, `gal` (warning: volumes expressed in
    USCS/imperial units are currently assumed to be US volumes)
    """

    WEIGHT = "weight"
    """Generic weight, represents a measurement of an object's mass.

    Weight is used instead of mass to fit with every day language.

    Unit of measurement: `MASS_*` units
    - SI / metric: `µg`, `mg`, `g`, `kg`
    - USCS / imperial: `oz`, `lb`
    """

    WIND_SPEED = "wind_speed"
    """Wind speed.

    Unit of measurement: `SPEED_*` units
    - SI /metric: `m/s`, `km/h`
    - USCS / imperial: `ft/s`, `mph`
    - Nautical: `kn`
    - Beaufort: `Beaufort`
    """


NON_NUMERIC_DEVICE_CLASSES = {
    SensorDeviceClass.DATE,
    SensorDeviceClass.ENUM,
    SensorDeviceClass.TIMESTAMP,
}

ANALOG_INPUT_APPTYPE_DEV_CLASS = {
    AnalogInputType.Temp_Degrees_C: SensorDeviceClass.TEMPERATURE,
    AnalogInputType.Relative_Humidity_Percent: SensorDeviceClass.HUMIDITY,
    AnalogInputType.Pressure_Pascal: SensorDeviceClass.PRESSURE,
    AnalogInputType.Flow_Liters_Per_Sec: SensorDeviceClass.VOLUME_FLOW_RATE,
    AnalogInputType.Percentage: None,
    AnalogInputType.Parts_Per_Million: None,  # It can be one of many
    AnalogInputType.Rotational_Speed_RPM: None,  # No device class for RPM
    AnalogInputType.Current_Amps: SensorDeviceClass.CURRENT,
    AnalogInputType.Frequency_Hz: SensorDeviceClass.FREQUENCY,
    AnalogInputType.Power_Watts: SensorDeviceClass.POWER,
    AnalogInputType.Power_Kilo_Watts: SensorDeviceClass.POWER,
    AnalogInputType.Energy_Kilo_Watt_Hours: SensorDeviceClass.ENERGY,
    AnalogInputType.Count: None,
    AnalogInputType.Enthalpy_KJoules_Per_Kg: None,
    AnalogInputType.Time_Seconds: SensorDeviceClass.DURATION,
}

ANALOG_INPUT_APPTYPE_UNITS = {
    AnalogInputType.Temp_Degrees_C: UnitOfTemperature.CELSIUS,
    AnalogInputType.Relative_Humidity_Percent: PERCENTAGE,
    AnalogInputType.Pressure_Pascal: UnitOfPressure.PA,
    AnalogInputType.Flow_Liters_Per_Sec: "L/s",
    AnalogInputType.Percentage: PERCENTAGE,
    AnalogInputType.Parts_Per_Million: CONCENTRATION_PARTS_PER_MILLION,
    AnalogInputType.Rotational_Speed_RPM: REVOLUTIONS_PER_MINUTE,
    AnalogInputType.Current_Amps: UnitOfElectricCurrent.AMPERE,
    AnalogInputType.Frequency_Hz: UnitOfFrequency.HERTZ,
    AnalogInputType.Power_Watts: UnitOfPower.WATT,
    AnalogInputType.Power_Kilo_Watts: UnitOfPower.KILO_WATT,
    AnalogInputType.Energy_Kilo_Watt_Hours: UnitOfEnergy.KILO_WATT_HOUR,
    AnalogInputType.Count: COUNT,
    AnalogInputType.Enthalpy_KJoules_Per_Kg: KILOJOULES_PER_KG,
    AnalogInputType.Time_Seconds: UnitOfTime.SECONDS,
}

ZCL_EPOCH = datetime(2000, 1, 1, tzinfo=UTC)



"""Support for ZHA sirens."""

from __future__ import annotations

import asyncio
import contextlib
from dataclasses import dataclass
from enum import IntFlag
import functools
from typing import TYPE_CHECKING, Any, Final, cast

from zigpy.profiles import zha
from zigpy.zcl.clusters.security import IasWd

from zha.application import Platform
from zha.application.const import (
    WARNING_DEVICE_MODE_BURGLAR,
    WARNING_DEVICE_MODE_EMERGENCY,
    WARNING_DEVICE_MODE_EMERGENCY_PANIC,
    WARNING_DEVICE_MODE_FIRE,
    WARNING_DEVICE_MODE_FIRE_PANIC,
    WARNING_DEVICE_MODE_POLICE_PANIC,
    WARNING_DEVICE_MODE_STOP,
    WARNING_DEVICE_SOUND_HIGH,
    WARNING_DEVICE_STROBE_HIGH,
    WARNING_DEVICE_STROBE_NO,
    Strobe,
)
from zha.application.platforms import (
    BaseEntityInfo,
    ClusterHandlerMatch,
    PlatformEntity,
    register_entity,
)
from zha.zigbee.cluster_handlers.const import CLUSTER_HANDLER_IAS_WD
from zha.zigbee.cluster_handlers.security import IasWdClusterHandler

if TYPE_CHECKING:
    from zha.zigbee.cluster_handlers import ClusterHandler
    from zha.zigbee.device import Device
    from zha.zigbee.endpoint import Endpoint

DEFAULT_DURATION = 5  # seconds

ATTR_AVAILABLE_TONES: Final[str] = "available_tones"
ATTR_DURATION: Final[str] = "duration"
ATTR_VOLUME_LEVEL: Final[str] = "volume_level"
ATTR_TONE: Final[str] = "tone"


class SirenEntityFeature(IntFlag):
    """Supported features of the siren entity."""

    TURN_ON = 1
    TURN_OFF = 2
    TONES = 4
    VOLUME_SET = 8
    DURATION = 16


@dataclass(frozen=True, kw_only=True)
class SirenEntityInfo(BaseEntityInfo):
    """Siren entity info."""

    available_tones: dict[int, str]
    supported_features: SirenEntityFeature


@register_entity(IasWd.cluster_id)
class Siren(PlatformEntity):
    """Representation of a ZHA siren."""

    PLATFORM = Platform.SIREN
    _attr_fallback_name: str = "Siren"
    _attr_primary_weight = 4

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_IAS_WD}),
    )

    def __init__(
        self,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> None:
        """Init this siren."""
        self._cluster_handler: IasWdClusterHandler = cast(
            IasWdClusterHandler, cluster_handlers[0]
        )

        legacy_discovery_unique_id = (
            f"{endpoint.device.ieee}-{endpoint.id}"
            if (
                endpoint.zigpy_endpoint.device_type == zha.DeviceType.IAS_WARNING_DEVICE
            )
            else f"{endpoint.device.ieee}-{endpoint.id}-{int(IasWd.cluster_id)}"
        )

        super().__init__(
            cluster_handlers,
            endpoint,
            device,
            **kwargs,
            legacy_discovery_unique_id=legacy_discovery_unique_id,
        )
        self._attr_supported_features = (
            SirenEntityFeature.TURN_ON
            | SirenEntityFeature.TURN_OFF
            | SirenEntityFeature.DURATION
            | SirenEntityFeature.VOLUME_SET
            | SirenEntityFeature.TONES
        )
        self._attr_available_tones: dict[int, str] = {
            WARNING_DEVICE_MODE_BURGLAR: "Burglar",
            WARNING_DEVICE_MODE_FIRE: "Fire",
            WARNING_DEVICE_MODE_EMERGENCY: "Emergency",
            WARNING_DEVICE_MODE_POLICE_PANIC: "Police Panic",
            WARNING_DEVICE_MODE_FIRE_PANIC: "Fire Panic",
            WARNING_DEVICE_MODE_EMERGENCY_PANIC: "Emergency Panic",
        }
        self._attr_is_on: bool = False
        self._off_listener: asyncio.TimerHandle | None = None

    @functools.cached_property
    def info_object(self) -> SirenEntityInfo:
        """Return representation of the siren."""
        return SirenEntityInfo(
            **super().info_object.__dict__,
            available_tones=self._attr_available_tones,
            supported_features=self._attr_supported_features,
        )

    @property
    def state(self) -> dict[str, Any]:
        """Get the state of the siren."""
        response = super().state
        response["state"] = self.is_on
        return response

    @property
    def supported_features(self) -> SirenEntityFeature:
        """Return supported features."""
        return self._attr_supported_features

    @property
    def is_on(self) -> bool:
        """Return true if the entity is on."""
        return self._attr_is_on

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn on siren."""
        if self._off_listener:
            self._off_listener.cancel()
            self._off_listener = None
        tone_cache = self._cluster_handler.data_cache.get(
            IasWd.Warning.WarningMode.__name__
        )
        siren_tone = (
            tone_cache.value
            if tone_cache is not None
            else WARNING_DEVICE_MODE_EMERGENCY
        )
        siren_duration = DEFAULT_DURATION
        level_cache = self._cluster_handler.data_cache.get(
            IasWd.Warning.SirenLevel.__name__
        )
        siren_level = (
            level_cache.value if level_cache is not None else WARNING_DEVICE_SOUND_HIGH
        )
        strobe_cache = self._cluster_handler.data_cache.get(Strobe.__name__)
        should_strobe = (
            strobe_cache.value if strobe_cache is not None else Strobe.No_Strobe
        )
        strobe_level_cache = self._cluster_handler.data_cache.get(
            IasWd.StrobeLevel.__name__
        )
        strobe_level = (
            strobe_level_cache.value
            if strobe_level_cache is not None
            else WARNING_DEVICE_STROBE_HIGH
        )
        if (duration := kwargs.get(ATTR_DURATION)) is not None:
            siren_duration = duration
        if (tone := kwargs.get(ATTR_TONE)) is not None:
            siren_tone = tone
        if (level := kwargs.get(ATTR_VOLUME_LEVEL)) is not None:
            siren_level = int(level)
        await self._cluster_handler.issue_start_warning(
            mode=siren_tone,
            warning_duration=siren_duration,
            siren_level=siren_level,
            strobe=should_strobe,
            strobe_duty_cycle=50 if should_strobe else 0,
            strobe_intensity=strobe_level,
        )
        self._attr_is_on = True
        self._off_listener = asyncio.get_running_loop().call_later(
            siren_duration, self.async_set_off
        )
        self._tracked_handles.append(self._off_listener)
        self.maybe_emit_state_changed_event()

    async def async_turn_off(self, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Turn off siren."""
        await self._cluster_handler.issue_start_warning(
            mode=WARNING_DEVICE_MODE_STOP, strobe=WARNING_DEVICE_STROBE_NO
        )
        self._attr_is_on = False
        self.maybe_emit_state_changed_event()

    def async_set_off(self) -> None:
        """Set is_on to False and write HA state."""
        self._attr_is_on = False
        if self._off_listener:
            self._off_listener.cancel()

            with contextlib.suppress(ValueError):
                self._tracked_handles.remove(self._off_listener)

            self._off_listener = None
        self.maybe_emit_state_changed_event()



"""Units of measure for Zigbee Home Automation."""

from enum import StrEnum
from typing import Final


class UnitOfTemperature(StrEnum):
    """Temperature units."""

    CELSIUS = "°C"
    FAHRENHEIT = "°F"
    KELVIN = "K"


class UnitOfMass(StrEnum):
    """Mass units."""

    GRAMS = "g"
    KILOGRAMS = "kg"
    MILLIGRAMS = "mg"
    MICROGRAMS = "µg"
    OUNCES = "oz"
    POUNDS = "lb"
    STONES = "st"


class UnitOfPressure(StrEnum):
    """Pressure units."""

    PA = "Pa"
    HPA = "hPa"
    KPA = "kPa"
    BAR = "bar"
    CBAR = "cbar"
    MBAR = "mbar"
    MMHG = "mmHg"
    INHG = "inHg"
    PSI = "psi"


class UnitOfSoundPressure(StrEnum):
    """Sound pressure units."""

    DECIBEL = "dB"
    WEIGHTED_DECIBEL_A = "dBA"


class UnitOfPower(StrEnum):
    """Power units."""

    MILLIWATT = "mW"
    WATT = "W"
    KILO_WATT = "kW"
    MEGA_WATT = "MW"
    GIGA_WATT = "GW"
    TERA_WATT = "TW"
    BTU_PER_HOUR = "BTU/h"


class UnitOfReactivePower(StrEnum):
    """Reactive power units."""

    VOLT_AMPERE_REACTIVE = "var"
    KILO_VOLT_AMPERE_REACTIVE = "kvar"


class UnitOfApparentPower(StrEnum):
    """Apparent power units."""

    VOLT_AMPERE = "VA"


class UnitOfElectricCurrent(StrEnum):
    """Electric current units."""

    MILLIAMPERE = "mA"
    AMPERE = "A"


# Electric_potential units
class UnitOfElectricPotential(StrEnum):
    """Electric potential units."""

    MICROVOLT = "µV"
    MILLIVOLT = "mV"
    VOLT = "V"
    KILOVOLT = "kV"
    MEGAVOLT = "MV"


class UnitOfFrequency(StrEnum):
    """Frequency units."""

    HERTZ = "Hz"
    KILOHERTZ = "kHz"
    MEGAHERTZ = "MHz"
    GIGAHERTZ = "GHz"


class UnitOfVolumeFlowRate(StrEnum):
    """Volume flow rate units."""

    CUBIC_METERS_PER_HOUR = "m³/h"
    CUBIC_METERS_PER_SECOND = "m³/s"
    CUBIC_FEET_PER_MINUTE = "ft³/min"
    LITERS_PER_HOUR = "L/h"
    LITERS_PER_MINUTE = "L/min"
    LITERS_PER_SECOND = "L/s"
    GALLONS_PER_MINUTE = "gal/min"
    MILLILITERS_PER_SECOND = "mL/s"


class UnitOfArea(StrEnum):
    """Area units."""

    SQUARE_METERS = "m²"
    SQUARE_CENTIMETERS = "cm²"
    SQUARE_KILOMETERS = "km²"
    SQUARE_MILLIMETERS = "mm²"
    SQUARE_INCHES = "in²"
    SQUARE_FEET = "ft²"
    SQUARE_YARDS = "yd²"
    SQUARE_MILES = "mi²"
    ACRES = "ac"
    HECTARES = "ha"


class UnitOfVolume(StrEnum):
    """Volume units."""

    CUBIC_FEET = "ft³"
    CENTUM_CUBIC_FEET = "CCF"
    CUBIC_METERS = "m³"
    LITERS = "L"
    MILLILITERS = "mL"
    GALLONS = "gal"
    """Assumed to be US gallons in conversion utilities.

    British/Imperial gallons are not yet supported"""
    FLUID_OUNCES = "fl. oz."
    """Assumed to be US fluid ounces in conversion utilities.

    British/Imperial fluid ounces are not yet supported"""


class UnitOfTime(StrEnum):
    """Time units."""

    MICROSECONDS = "μs"
    MILLISECONDS = "ms"
    SECONDS = "s"
    MINUTES = "min"
    HOURS = "h"
    DAYS = "d"
    WEEKS = "w"
    MONTHS = "m"
    YEARS = "y"


class UnitOfLength(StrEnum):
    """Length units."""

    MILLIMETERS = "mm"
    CENTIMETERS = "cm"
    METERS = "m"
    KILOMETERS = "km"
    INCHES = "in"
    FEET = "ft"
    YARDS = "yd"
    MILES = "mi"
    NAUTICAL_MILES = "nmi"


class UnitOfEnergy(StrEnum):
    """Energy units."""

    JOULE = "J"
    KILO_JOULE = "kJ"
    MEGA_JOULE = "MJ"
    GIGA_JOULE = "GJ"
    MILLIWATT_HOUR = "mWh"
    WATT_HOUR = "Wh"
    KILO_WATT_HOUR = "kWh"
    MEGA_WATT_HOUR = "MWh"
    GIGA_WATT_HOUR = "GWh"
    TERA_WATT_HOUR = "TWh"
    CALORIE = "cal"
    KILO_CALORIE = "kcal"
    MEGA_CALORIE = "Mcal"
    GIGA_CALORIE = "Gcal"


class UnitOfEnergyDistance(StrEnum):
    """Energy Distance units."""

    KILO_WATT_HOUR_PER_100_KM = "kWh/100km"
    MILES_PER_KILO_WATT_HOUR = "mi/kWh"
    KM_PER_KILO_WATT_HOUR = "km/kWh"


class UnitOfConductivity(StrEnum):
    """Conductivity units."""

    SIEMENS_PER_CM = "S/cm"
    MICROSIEMENS_PER_CM = "µS/cm"
    MILLISIEMENS_PER_CM = "mS/cm"


class UnitOfSpeed(StrEnum):
    """Speed units."""

    BEAUFORT = "Beaufort"
    FEET_PER_SECOND = "ft/s"
    INCHES_PER_SECOND = "in/s"
    METERS_PER_SECOND = "m/s"
    KILOMETERS_PER_HOUR = "km/h"
    KNOTS = "kn"
    MILES_PER_HOUR = "mph"
    MILLIMETERS_PER_SECOND = "mm/s"


# Concentration units
CONCENTRATION_MICROGRAMS_PER_CUBIC_METER: Final = "µg/m³"
CONCENTRATION_MILLIGRAMS_PER_CUBIC_METER: Final = "mg/m³"
CONCENTRATION_MICROGRAMS_PER_CUBIC_FOOT: Final = "μg/ft³"
CONCENTRATION_PARTS_PER_CUBIC_METER: Final = "p/m³"
CONCENTRATION_PARTS_PER_MILLION: Final = "ppm"
CONCENTRATION_PARTS_PER_BILLION: Final = "ppb"

# Signal_strength units
SIGNAL_STRENGTH_DECIBELS: Final = "dB"
SIGNAL_STRENGTH_DECIBELS_MILLIWATT: Final = "dBm"

# Light units
LIGHT_LUX: Final = "lx"

# Percentage units
PERCENTAGE: Final[str] = "%"

# Rotational speed units
REVOLUTIONS_PER_MINUTE: Final = "rpm"


# Irradiance units
class UnitOfIrradiance(StrEnum):
    """Irradiance units."""

    WATTS_PER_SQUARE_METER = "W/m²"
    BTUS_PER_HOUR_SQUARE_FOOT = "BTU/(h⋅ft²)"


# Kinetic energy
KILOJOULES_PER_KG: Final = "KJ/kg"

# Count
COUNT: Final = "count"

DEGREE: Final = "°"



"""Constants for the Light platform."""

from enum import IntFlag, StrEnum
from typing import Final

from zigpy.profiles import zha, zll
from zigpy.zcl.clusters.general import Identify

DEFAULT_ON_OFF_TRANSITION = 1  # most bulbs default to a 1-second turn on/off transition
DEFAULT_EXTRA_TRANSITION_DELAY_SHORT = 0.25
DEFAULT_EXTRA_TRANSITION_DELAY_LONG = 2.0
DEFAULT_LONG_TRANSITION_TIME = 10
DEFAULT_MIN_BRIGHTNESS = 2
ASSUME_UPDATE_GROUP_FROM_CHILD_DELAY = 0.05

DEFAULT_MIN_TRANSITION_MANUFACTURERS = {"sengled"}

STATE_UNAVAILABLE: Final[str] = "unavailable"


class LightEntityFeature(IntFlag):
    """Supported features of the light entity."""

    EFFECT = 4
    FLASH = 8
    TRANSITION = 32


class ColorMode(StrEnum):
    """Possible light color modes."""

    UNKNOWN = "unknown"
    ONOFF = "onoff"
    BRIGHTNESS = "brightness"
    COLOR_TEMP = "color_temp"
    XY = "xy"


# Float that represents transition time in seconds to make change.
ATTR_TRANSITION: Final[str] = "transition"

# Lists holding color values
ATTR_XY_COLOR: Final[str] = "xy_color"
ATTR_COLOR_TEMP: Final[str] = "color_temp"
ATTR_MIN_MIREDS: Final[str] = "min_mireds"
ATTR_MAX_MIREDS: Final[str] = "max_mireds"

# Brightness of the light, 0..255 or percentage
ATTR_BRIGHTNESS: Final[str] = "brightness"

ATTR_COLOR_MODE = "color_mode"
ATTR_SUPPORTED_COLOR_MODES = "supported_color_modes"

# If the light should flash, can be FLASH_SHORT or FLASH_LONG.
ATTR_FLASH: Final[str] = "flash"


class FlashMode(StrEnum):
    """Flash modes."""

    SHORT = "short"
    LONG = "long"


# List of possible effects
ATTR_EFFECT_LIST: Final[str] = "effect_list"

# Apply an effect to the light, can be EFFECT_COLORLOOP.
ATTR_EFFECT: Final[str] = "effect"
EFFECT_COLORLOOP: Final[str] = "colorloop"
EFFECT_RANDOM: Final[str] = "random"
EFFECT_WHITE: Final[str] = "white"
EFFECT_OFF: Final[str] = "off"

ATTR_SUPPORTED_FEATURES: Final[str] = "supported_features"

EFFECT_OKAY: Final[int] = 0x02
EFFECT_DEFAULT_VARIANT: Final[int] = 0x00

FLASH_EFFECTS: Final[dict[str, int]] = {
    FlashMode.SHORT: Identify.EffectIdentifier.Blink,
    FlashMode.LONG: Identify.EffectIdentifier.Breathe,
}

VALID_COLOR_MODES = {
    ColorMode.ONOFF,
    ColorMode.BRIGHTNESS,
    ColorMode.COLOR_TEMP,
    ColorMode.XY,
}
COLOR_MODES_BRIGHTNESS = VALID_COLOR_MODES - {ColorMode.ONOFF}
COLOR_MODES_COLOR = {ColorMode.XY}

LIGHT_PROFILE_DEVICE_TYPES = frozenset(
    {
        # ZHA
        (zha.PROFILE_ID, zha.DeviceType.COLOR_DIMMABLE_LIGHT),
        (zha.PROFILE_ID, zha.DeviceType.COLOR_TEMPERATURE_LIGHT),
        (zha.PROFILE_ID, zha.DeviceType.DIMMABLE_BALLAST),
        (zha.PROFILE_ID, zha.DeviceType.DIMMABLE_LIGHT),
        (zha.PROFILE_ID, zha.DeviceType.DIMMABLE_PLUG_IN_UNIT),
        (zha.PROFILE_ID, zha.DeviceType.EXTENDED_COLOR_LIGHT),
        (zha.PROFILE_ID, zha.DeviceType.ON_OFF_LIGHT),
        # ZLL
        (zll.PROFILE_ID, zll.DeviceType.COLOR_LIGHT),
        (zll.PROFILE_ID, zll.DeviceType.COLOR_TEMPERATURE_LIGHT),
        (zll.PROFILE_ID, zll.DeviceType.DIMMABLE_LIGHT),
        (zll.PROFILE_ID, zll.DeviceType.DIMMABLE_PLUGIN_UNIT),
        (zll.PROFILE_ID, zll.DeviceType.EXTENDED_COLOR_LIGHT),
        (zll.PROFILE_ID, zll.DeviceType.ON_OFF_LIGHT),
    }
)
